import datetime
import logging
import random
import string

import salt.loader
import salt.loader.context
import salt.modules.boto_apigateway as boto_apigateway
from salt.utils.versions import LooseVersion
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

# pylint: disable=import-error,no-name-in-module
try:
    import boto3
    import botocore
    from botocore.exceptions import ClientError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False


# pylint: enable=import-error,no-name-in-module

# the boto_lambda module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = "1.2.1"
required_botocore_version = "1.4.49"

region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
error_message = (
    "An error occurred (101) when calling the {0} operation: Test-defined error"
)
error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}

api_key_ret = {
    "description": "test-lambda-api-key",
    "enabled": True,
    "stageKeys": ["123yd1l123/test"],
    "lastUpdatedDate": datetime.datetime(2015, 11, 4, 19, 22, 18),
    "createdDate": datetime.datetime(2015, 11, 4, 19, 21, 7),
    "id": "88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
    "name": "test-salt-key",
    "ResponseMetadata": {
        "HTTPStatusCode": 200,
        "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
    },
}

api_model_error_schema = '{"properties":{"code":{"type":"integer","format":"int32"},"message":{"type":"string"},"fields":{"type":"string"}},"definitions":{}}'
api_model_ret = {
    "contentType": "application/json",
    "name": "Error",
    "description": "Error Model",
    "id": "iltqcb",
    "schema": api_model_error_schema,
    "ResponseMetadata": {
        "HTTPStatusCode": 200,
        "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
    },
}

api_resources_ret = {
    "items": [
        {"id": "hhg2t8f4h9", "path": "/"},
        {"id": "isr8q2", "parentId": "hhg2t8f4h9", "path": "/api", "pathPart": "api"},
        {
            "id": "5pvx7w",
            "parentId": "isr8q2",
            "path": "/api/users",
            "pathPart": "users",
            "resourceMethods": {"OPTIONS": {}, "POST": {}},
        },
    ],
    "ResponseMetadata": {
        "HTTPStatusCode": 200,
        "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
    },
}

api_create_resource_ret = {
    "id": "123abc",
    "parentId": "hhg2t8f4h9",
    "path": "/api3",
    "pathPart": "api3",
    "ResponseMetadata": {
        "HTTPStatusCode": 200,
        "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
    },
}

usage_plan1 = dict(
    id="plan1_id",
    name="plan1_name",
    description="plan1_desc",
    apiStages=[],
    throttle=dict(burstLimit=123, rateLimit=123.0),
    quota=dict(limit=123, offset=123, period="DAY"),
)
usage_plan2 = dict(
    id="plan2_id",
    name="plan2_name",
    description="plan2_desc",
    apiStages=[],
    throttle=dict(burstLimit=123, rateLimit=123.0),
    quota=dict(limit=123, offset=123, period="DAY"),
)
usage_plan1b = dict(
    id="another_plan1_id",
    name="plan1_name",
    description="another_plan1_desc",
    apiStages=[],
    throttle=dict(burstLimit=123, rateLimit=123.0),
    quota=dict(limit=123, offset=123, period="DAY"),
)
usage_plans_ret = dict(items=[usage_plan1, usage_plan2, usage_plan1b])

log = logging.getLogger(__name__)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif LooseVersion(boto3.__version__) < LooseVersion(required_boto3_version):
        return False
    else:
        return True


def _has_required_botocore():
    """
    Returns True/False boolean depending on if botocore supports usage plan
    """
    if not HAS_BOTO:
        return False
    elif LooseVersion(botocore.__version__) < LooseVersion(required_botocore_version):
        return False
    else:
        return True


class BotoApiGatewayTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn = None

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        self.utils = salt.loader.utils(
            opts, whitelist=["boto3", "args", "systemd", "path", "platform"]
        )
        return {boto_apigateway: {"__opts__": opts, "__utils__": self.utils}}

    def setUp(self):
        TestCase.setUp(self)
        # __virtual__ must be caller in order for _get_conn to be injected
        boto_apigateway.__init__(self.opts)
        delattr(self, "opts")

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        session_instance.client.return_value = self.conn
        self.addCleanup(delattr, self, "conn")
        self.addCleanup(delattr, self, "utils")


class BotoApiGatewayTestCaseMixin:
    def _diff_list_dicts(self, listdict1, listdict2, sortkey):
        """
        Compares the two list of dictionaries to ensure they have same content.  Returns True
        if there is difference, else False
        """
        if len(listdict1) != len(listdict2):
            return True

        listdict1_sorted = sorted(listdict1, key=lambda x: x[sortkey])
        listdict2_sorted = sorted(listdict2, key=lambda x: x[sortkey])
        for item1, item2 in zip(listdict1_sorted, listdict2_sorted):
            if len(set(item1) & set(item2)) != len(set(item2)):
                return True
        return False


# @skipIf(True, 'Skip these tests while investigating failures')
@skipIf(HAS_BOTO is False, "The boto3 module must be installed.")
@skipIf(
    _has_required_boto() is False,
    "The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
@skipIf(
    _has_required_botocore() is False,
    "The botocore module must be greater than or equal to version {}".format(
        required_botocore_version
    ),
)
class BotoApiGatewayTestCase(BotoApiGatewayTestCaseBase, BotoApiGatewayTestCaseMixin):
    """
    TestCase for salt.modules.boto_apigateway module
    """

    def test_that_when_checking_if_a_rest_api_exists_and_a_rest_api_exists_the_api_exists_method_returns_true(
        self,
    ):
        """
        Tests checking an apigateway rest api existence when api's name exists
        """
        self.conn.get_rest_apis.return_value = {
            "items": [{"name": "myapi", "id": "1234def"}]
        }
        api_exists_result = boto_apigateway.api_exists(name="myapi", **conn_parameters)

        self.assertTrue(api_exists_result["exists"])

    def test_that_when_checking_if_a_rest_api_exists_and_multiple_rest_api_exist_the_api_exists_method_returns_true(
        self,
    ):
        """
        Tests checking an apigateway rest api existence when multiple api's with same name exists
        """
        self.conn.get_rest_apis.return_value = {
            "items": [
                {"name": "myapi", "id": "1234abc"},
                {"name": "myapi", "id": "1234def"},
            ]
        }
        api_exists_result = boto_apigateway.api_exists(name="myapi", **conn_parameters)
        self.assertTrue(api_exists_result["exists"])

    def test_that_when_checking_if_a_rest_api_exists_and_no_rest_api_exists_the_api_exists_method_returns_false(
        self,
    ):
        """
        Tests checking an apigateway rest api existence when no matching rest api name exists
        """
        self.conn.get_rest_apis.return_value = {
            "items": [
                {"name": "myapi", "id": "1234abc"},
                {"name": "myapi", "id": "1234def"},
            ]
        }
        api_exists_result = boto_apigateway.api_exists(
            name="myapi123", **conn_parameters
        )
        self.assertFalse(api_exists_result["exists"])

    def test_that_when_describing_rest_apis_and_no_name_given_the_describe_apis_method_returns_list_of_all_rest_apis(
        self,
    ):
        """
        Tests that all rest apis defined for a region is returned
        """
        self.conn.get_rest_apis.return_value = {
            "items": [
                {
                    "description": (
                        "A sample API that uses a petstore as an example to demonstrate"
                        " features in the swagger-2.0 specification"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
                    "id": "2ut6i4vyle",
                    "name": "Swagger Petstore",
                },
                {
                    "description": "testingabcd",
                    "createdDate": datetime.datetime(2015, 12, 3, 21, 57, 58),
                    "id": "g41ls77hz0",
                    "name": "testingabc",
                },
                {
                    "description": "a simple food delivery service test",
                    "createdDate": datetime.datetime(2015, 11, 4, 23, 57, 28),
                    "id": "h7pbwydho9",
                    "name": "Food Delivery Service",
                },
                {
                    "description": "Created by AWS Lambda",
                    "createdDate": datetime.datetime(2015, 11, 4, 17, 55, 41),
                    "id": "i2yyd1ldvj",
                    "name": "LambdaMicroservice",
                },
                {
                    "description": (
                        "cloud tap service with combination of API GW and Lambda"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 22, 3, 18),
                    "id": "rm06h9oac4",
                    "name": "API Gateway Cloudtap Service",
                },
                {
                    "description": "testing1234",
                    "createdDate": datetime.datetime(2015, 12, 2, 19, 51, 44),
                    "id": "vtir6ssxvd",
                    "name": "testing123",
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        items = self.conn.get_rest_apis.return_value["items"]
        get_apis_result = boto_apigateway.describe_apis(**conn_parameters)
        items_dt = [boto_apigateway._convert_datetime_str(item) for item in items]
        apis = get_apis_result.get("restapi")

        diff = self._diff_list_dicts(apis, items_dt, "id")

        self.assertTrue(apis)
        self.assertEqual(len(apis), len(items))
        self.assertFalse(diff)

    def test_that_when_describing_rest_apis_and_name_is_testing123_the_describe_apis_method_returns_list_of_two_rest_apis(
        self,
    ):
        """
        Tests that exactly 2 apis are returned matching 'testing123'
        """
        self.conn.get_rest_apis.return_value = {
            "items": [
                {
                    "description": (
                        "A sample API that uses a petstore as an example to demonstrate"
                        " features in the swagger-2.0 specification"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
                    "id": "2ut6i4vyle",
                    "name": "Swagger Petstore",
                },
                {
                    "description": "testingabcd",
                    "createdDate": datetime.datetime(2015, 12, 3, 21, 57, 58),
                    "id": "g41ls77hz0",
                    "name": "testing123",
                },
                {
                    "description": "a simple food delivery service test",
                    "createdDate": datetime.datetime(2015, 11, 4, 23, 57, 28),
                    "id": "h7pbwydho9",
                    "name": "Food Delivery Service",
                },
                {
                    "description": "Created by AWS Lambda",
                    "createdDate": datetime.datetime(2015, 11, 4, 17, 55, 41),
                    "id": "i2yyd1ldvj",
                    "name": "LambdaMicroservice",
                },
                {
                    "description": (
                        "cloud tap service with combination of API GW and Lambda"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 22, 3, 18),
                    "id": "rm06h9oac4",
                    "name": "API Gateway Cloudtap Service",
                },
                {
                    "description": "testing1234",
                    "createdDate": datetime.datetime(2015, 12, 2, 19, 51, 44),
                    "id": "vtir6ssxvd",
                    "name": "testing123",
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        expected_items = [
            {
                "description": "testingabcd",
                "createdDate": datetime.datetime(2015, 12, 3, 21, 57, 58),
                "id": "g41ls77hz0",
                "name": "testing123",
            },
            {
                "description": "testing1234",
                "createdDate": datetime.datetime(2015, 12, 2, 19, 51, 44),
                "id": "vtir6ssxvd",
                "name": "testing123",
            },
        ]

        get_apis_result = boto_apigateway.describe_apis(
            name="testing123", **conn_parameters
        )
        expected_items_dt = [
            boto_apigateway._convert_datetime_str(item) for item in expected_items
        ]
        apis = get_apis_result.get("restapi")
        diff = self._diff_list_dicts(apis, expected_items_dt, "id")

        self.assertTrue(apis)
        self.assertIs(diff, False)

    def test_that_when_describing_rest_apis_and_name_is_testing123_the_describe_apis_method_returns_no_matching_items(
        self,
    ):
        """
        Tests that no apis are returned matching 'testing123'
        """
        self.conn.get_rest_apis.return_value = {
            "items": [
                {
                    "description": (
                        "A sample API that uses a petstore as an example to demonstrate"
                        " features in the swagger-2.0 specification"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
                    "id": "2ut6i4vyle",
                    "name": "Swagger Petstore",
                },
                {
                    "description": "a simple food delivery service test",
                    "createdDate": datetime.datetime(2015, 11, 4, 23, 57, 28),
                    "id": "h7pbwydho9",
                    "name": "Food Delivery Service",
                },
                {
                    "description": "Created by AWS Lambda",
                    "createdDate": datetime.datetime(2015, 11, 4, 17, 55, 41),
                    "id": "i2yyd1ldvj",
                    "name": "LambdaMicroservice",
                },
                {
                    "description": (
                        "cloud tap service with combination of API GW and Lambda"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 22, 3, 18),
                    "id": "rm06h9oac4",
                    "name": "API Gateway Cloudtap Service",
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        get_apis_result = boto_apigateway.describe_apis(
            name="testing123", **conn_parameters
        )
        apis = get_apis_result.get("restapi")
        self.assertFalse(apis)

    def test_that_when_creating_a_rest_api_succeeds_the_create_api_method_returns_true(
        self,
    ):
        """
        test True if rest api is created
        """
        created_date = datetime.datetime.now()
        assigned_api_id = "created_api_id"
        self.conn.create_rest_api.return_value = {
            "description": "unit-testing1234",
            "createdDate": created_date,
            "id": assigned_api_id,
            "name": "unit-testing123",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }

        create_api_result = boto_apigateway.create_api(
            name="unit-testing123", description="unit-testing1234", **conn_parameters
        )
        api = create_api_result.get("restapi")
        self.assertTrue(create_api_result.get("created"))
        self.assertTrue(api)
        self.assertEqual(api["id"], assigned_api_id)
        self.assertEqual(api["createdDate"], "{}".format(created_date))
        self.assertEqual(api["name"], "unit-testing123")
        self.assertEqual(api["description"], "unit-testing1234")

    def test_that_when_creating_a_rest_api_fails_the_create_api_method_returns_error(
        self,
    ):
        """
        test True for rest api creation error.
        """
        self.conn.create_rest_api.side_effect = ClientError(
            error_content, "create_rest_api"
        )
        create_api_result = boto_apigateway.create_api(
            name="unit-testing123", description="unit-testing1234", **conn_parameters
        )
        api = create_api_result.get("restapi")
        self.assertEqual(
            create_api_result.get("error").get("message"),
            error_message.format("create_rest_api"),
        )

    def test_that_when_deleting_rest_apis_and_name_is_testing123_matching_two_apis_the_delete_api_method_returns_delete_count_of_two(
        self,
    ):
        """
        test True if the deleted count for "testing123" api is 2.
        """
        self.conn.get_rest_apis.return_value = {
            "items": [
                {
                    "description": (
                        "A sample API that uses a petstore as an example to demonstrate"
                        " features in the swagger-2.0 specification"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
                    "id": "2ut6i4vyle",
                    "name": "Swagger Petstore",
                },
                {
                    "description": "testingabcd",
                    "createdDate": datetime.datetime(2015, 12, 3, 21, 57, 58),
                    "id": "g41ls77hz0",
                    "name": "testing123",
                },
                {
                    "description": "a simple food delivery service test",
                    "createdDate": datetime.datetime(2015, 11, 4, 23, 57, 28),
                    "id": "h7pbwydho9",
                    "name": "Food Delivery Service",
                },
                {
                    "description": "Created by AWS Lambda",
                    "createdDate": datetime.datetime(2015, 11, 4, 17, 55, 41),
                    "id": "i2yyd1ldvj",
                    "name": "LambdaMicroservice",
                },
                {
                    "description": (
                        "cloud tap service with combination of API GW and Lambda"
                    ),
                    "createdDate": datetime.datetime(2015, 11, 17, 22, 3, 18),
                    "id": "rm06h9oac4",
                    "name": "API Gateway Cloudtap Service",
                },
                {
                    "description": "testing1234",
                    "createdDate": datetime.datetime(2015, 12, 2, 19, 51, 44),
                    "id": "vtir6ssxvd",
                    "name": "testing123",
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        self.conn.delete_rest_api.return_value = None
        delete_api_result = boto_apigateway.delete_api(
            name="testing123", **conn_parameters
        )

        self.assertTrue(delete_api_result.get("deleted"))
        self.assertEqual(delete_api_result.get("count"), 2)

    def test_that_when_deleting_rest_apis_and_name_given_provides_no_match_the_delete_api_method_returns_false(
        self,
    ):
        """
        Test that the given api name doesn't exists, and delete_api should return deleted status of False
        """
        self.conn.get_rest_apis.return_value = {
            "items": [
                {
                    "description": "testing1234",
                    "createdDate": datetime.datetime(2015, 12, 2, 19, 51, 44),
                    "id": "vtir6ssxvd",
                    "name": "testing1234",
                }
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        self.conn.delete_rest_api.return_value = None
        delete_api_result = boto_apigateway.delete_api(
            name="testing123", **conn_parameters
        )

        self.assertFalse(delete_api_result.get("deleted"))

    def test_that_describing_api_keys_the_describe_api_keys_method_returns_all_api_keys(
        self,
    ):
        """
        tests True if all api_keys are returned.
        """
        self.conn.get_api_keys.return_value = {
            "items": [
                {
                    "description": "test-lambda-api-key",
                    "enabled": True,
                    "stageKeys": ["123yd1l123/test"],
                    "lastUpdatedDate": datetime.datetime(2015, 11, 4, 19, 22, 18),
                    "createdDate": datetime.datetime(2015, 11, 4, 19, 21, 7),
                    "id": "88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
                    "name": "test-salt-key",
                },
                {
                    "description": "testing_salt_123",
                    "enabled": True,
                    "stageKeys": [],
                    "lastUpdatedDate": datetime.datetime(2015, 12, 5, 0, 14, 49),
                    "createdDate": datetime.datetime(2015, 12, 4, 22, 29, 33),
                    "id": "999999989b8cNSp4505pL6OgDe3oW7oY29Z3eIZ4",
                    "name": "testing_salt",
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }

        items = self.conn.get_api_keys.return_value["items"]
        get_api_keys_result = boto_apigateway.describe_api_keys(**conn_parameters)
        items_dt = [boto_apigateway._convert_datetime_str(item) for item in items]
        api_keys = get_api_keys_result.get("apiKeys")

        diff = False
        if len(api_keys) != len(items):
            diff = True
        else:
            # compare individual items.
            diff = self._diff_list_dicts(api_keys, items_dt, "id")

        self.assertTrue(api_keys)
        self.assertIs(diff, False)

    def test_that_describing_api_keys_fails_the_desribe_api_keys_method_returns_error(
        self,
    ):
        """
        test True for describe api keys error.
        """
        self.conn.get_api_keys.side_effect = ClientError(error_content, "get_api_keys")
        result = boto_apigateway.describe_api_keys(**conn_parameters)
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("get_api_keys")
        )

    def test_that_describing_an_api_key_the_describe_api_key_method_returns_matching_api_key(
        self,
    ):
        """
        tests True if the key is found.
        """
        self.conn.get_api_key.return_value = api_key_ret
        result = boto_apigateway.describe_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )
        self.assertEqual(
            result.get("apiKey", {}).get("id"),
            self.conn.get_api_key.return_value.get("id"),
        )

    def test_that_describing_an_api_key_that_does_not_exists_the_desribe_api_key_method_returns_error(
        self,
    ):
        """
        test True for error being thrown.
        """
        self.conn.get_api_key.side_effect = ClientError(error_content, "get_api_keys")
        result = boto_apigateway.describe_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("get_api_keys")
        )

    def test_that_when_creating_an_api_key_succeeds_the_create_api_key_method_returns_true(
        self,
    ):
        """
        tests that we can successfully create an api key and the createDat and lastUpdateDate are
        converted to string
        """
        now = datetime.datetime.now()
        self.conn.create_api_key.return_value = {
            "description": "test-lambda-api-key",
            "enabled": True,
            "stageKeys": ["123yd1l123/test"],
            "lastUpdatedDate": now,
            "createdDate": now,
            "id": "88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
            "name": "test-salt-key",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }

        create_api_key_result = boto_apigateway.create_api_key(
            "test-salt-key", "test-lambda-api-key", **conn_parameters
        )
        api_key = create_api_key_result.get("apiKey")
        now_str = "{}".format(now)

        self.assertTrue(create_api_key_result.get("created"))
        self.assertEqual(api_key.get("lastUpdatedDate"), now_str)
        self.assertEqual(api_key.get("createdDate"), now_str)

    def test_that_when_creating_an_api_key_fails_the_create_api_key_method_returns_error(
        self,
    ):
        """
        tests that we properly handle errors when create an api key fails.
        """
        with salt.loader.context.loader_context(self.utils):
            self.conn.create_api_key.side_effect = ClientError(
                error_content, "create_api_key"
            )
            create_api_key_result = boto_apigateway.create_api_key(
                "test-salt-key", "unit-testing1234"
            )
            api_key = create_api_key_result.get("apiKey")

            self.assertFalse(api_key)
            self.assertIs(create_api_key_result.get("created"), False)
            self.assertEqual(
                create_api_key_result.get("error").get("message"),
                error_message.format("create_api_key"),
            )

    def test_that_when_deleting_an_api_key_that_exists_the_delete_api_key_method_returns_true(
        self,
    ):
        """
        test True if the api key is successfully deleted.
        """
        self.conn.delete_api_key.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            }
        }
        result = boto_apigateway.delete_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )

        self.assertTrue(result.get("deleted"))

    def test_that_when_deleting_an_api_key_that_does_not_exist_the_delete_api_key_method_returns_false(
        self,
    ):
        """
        Test that the given api key doesn't exists, and delete_api_key should return deleted status of False
        """
        self.conn.delete_api_key.side_effect = ClientError(
            error_content, "delete_api_key"
        )
        result = boto_apigateway.delete_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )

        self.assertFalse(result.get("deleted"))

    def test_that_when_updating_an_api_key_description_successfully_the_update_api_key_description_method_returns_true(
        self,
    ):
        """
        Test True if api key descriptipn update is successful
        """
        self.conn.update_api_key.return_value = api_key_ret
        result = boto_apigateway.update_api_key_description(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
            description="test-lambda-api-key",
            **conn_parameters
        )
        self.assertTrue(result.get("updated"))

    def test_that_when_updating_an_api_key_description_for_a_key_that_does_not_exist_the_update_api_key_description_method_returns_false(
        self,
    ):
        """
        Test False if api key doesn't exists for the update request
        """
        self.conn.update_api_key.side_effect = ClientError(
            error_content, "update_api_key"
        )
        result = boto_apigateway.update_api_key_description(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
            description="test-lambda-api-key",
            **conn_parameters
        )
        self.assertFalse(result.get("updated"))

    def test_that_when_enabling_an_api_key_that_exists_the_enable_api_key_method_returns_api_key(
        self,
    ):
        """
        Test True for the status of the enabled flag of the returned api key
        """
        self.conn.update_api_key.return_value = api_key_ret
        result = boto_apigateway.enable_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )
        self.assertTrue(result.get("apiKey", {}).get("enabled"))

    def test_that_when_enabling_an_api_key_that_does_not_exist_the_enable_api_key_method_returns_error(
        self,
    ):
        """
        Test Equality of the returned value of 'erorr'
        """
        self.conn.update_api_key.side_effect = ClientError(
            error_content, "update_api_key"
        )
        result = boto_apigateway.enable_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("update_api_key")
        )

    def test_that_when_disabling_an_api_key_that_exists_the_disable_api_key_method_returns_api_key(
        self,
    ):
        """
        Test False for the status of the enabled flag of the returned api key
        """
        self.conn.update_api_key.return_value = api_key_ret.copy()
        self.conn.update_api_key.return_value["enabled"] = False
        result = boto_apigateway.disable_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )
        self.assertFalse(result.get("apiKey", {}).get("enabled"))

    def test_that_when_disabling_an_api_key_that_does_not_exist_the_disable_api_key_method_returns_error(
        self,
    ):
        """
        Test Equality of the returned value of 'erorr'
        """
        self.conn.update_api_key.side_effect = ClientError(
            error_content, "update_api_key"
        )
        result = boto_apigateway.disable_api_key(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("update_api_key")
        )

    def test_that_when_associating_stages_to_an_api_key_that_exists_the_associate_api_key_stagekeys_method_returns_true(
        self,
    ):
        """
        Test True for returned value of 'associated'
        """
        self.conn.update_api_key.retuen_value = api_key_ret
        result = boto_apigateway.associate_api_key_stagekeys(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
            stagekeyslist=["123yd1l123/test"],
            **conn_parameters
        )
        self.assertTrue(result.get("associated"))

    def test_that_when_associating_stages_to_an_api_key_that_does_not_exist_the_associate_api_key_stagekeys_method_returns_false(
        self,
    ):
        """
        Test False returned value of 'associated'
        """
        self.conn.update_api_key.side_effect = ClientError(
            error_content, "update_api_key"
        )
        result = boto_apigateway.associate_api_key_stagekeys(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
            stagekeyslist=["123yd1l123/test"],
            **conn_parameters
        )
        self.assertFalse(result.get("associated"))

    def test_that_when_disassociating_stages_to_an_api_key_that_exists_the_disassociate_api_key_stagekeys_method_returns_true(
        self,
    ):
        """
        Test True for returned value of 'associated'
        """
        self.conn.update_api_key.retuen_value = None
        result = boto_apigateway.disassociate_api_key_stagekeys(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
            stagekeyslist=["123yd1l123/test"],
            **conn_parameters
        )
        self.assertTrue(result.get("disassociated"))

    def test_that_when_disassociating_stages_to_an_api_key_that_does_not_exist_the_disassociate_api_key_stagekeys_method_returns_false(
        self,
    ):
        """
        Test False returned value of 'associated'
        """
        self.conn.update_api_key.side_effect = ClientError(
            error_content, "update_api_key"
        )
        result = boto_apigateway.disassociate_api_key_stagekeys(
            apiKey="88883333amaa1ZMVGCoLeaTrQk8kzOC36vCgRcT2",
            stagekeyslist=["123yd1l123/test"],
            **conn_parameters
        )
        self.assertFalse(result.get("disassociated"))

    def test_that_when_describing_api_deployments_the_describe_api_deployments_method_returns_list_of_deployments(
        self,
    ):
        """
        Test Equality for number of deployments is 2
        """
        self.conn.get_deployments.return_value = {
            "items": [
                {
                    "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
                    "id": "n05smo",
                },
                {
                    "createdDate": datetime.datetime(2015, 12, 2, 19, 51, 44),
                    "id": "n05sm1",
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        result = boto_apigateway.describe_api_deployments(
            restApiId="rm06h9oac4", **conn_parameters
        )
        self.assertEqual(len(result.get("deployments", {})), 2)

    def test_that_when_describing_api_deployments_and_an_error_occurred_the_describe_api_deployments_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_deployments.side_effect = ClientError(
            error_content, "get_deployments"
        )
        result = boto_apigateway.describe_api_deployments(
            restApiId="rm06h9oac4", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_deployments")
        )

    def test_that_when_describing_an_api_deployment_the_describe_api_deployment_method_returns_the_deployment(
        self,
    ):
        """
        Test True for the returned deployment
        """
        self.conn.get_deployment.return_value = {
            "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "id": "n05smo",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        result = boto_apigateway.describe_api_deployment(
            restApiId="rm06h9oac4", deploymentId="n05smo", **conn_parameters
        )
        self.assertTrue(result.get("deployment"))

    def test_that_when_describing_api_deployment_that_does_not_exist_the_describe_api_deployment_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_deployment.side_effect = ClientError(
            error_content, "get_deployment"
        )
        result = boto_apigateway.describe_api_deployment(
            restApiId="rm06h9oac4", deploymentId="n05smo", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_deployment")
        )

    def test_that_when_activating_api_deployment_for_stage_and_deployment_that_exist_the_activate_api_deployment_method_returns_true(
        self,
    ):
        """
        Test True for value of 'set'
        """
        self.conn.update_stage.return_value = {
            "cacheClusterEnabled": False,
            "cacheClusterStatus": "NOT_AVAAILABLE",
            "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "deploymentId": "n05smo",
            "description": "test",
            "lastUpdatedDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "stageName": "test",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        result = boto_apigateway.activate_api_deployment(
            restApiId="rm06h9oac4",
            stageName="test",
            deploymentId="n05smo",
            **conn_parameters
        )
        self.assertTrue(result.get("set"))

    def test_that_when_activating_api_deployment_for_stage_that_does_not_exist_the_activate_api_deployment_method_returns_false(
        self,
    ):
        """
        Test False for value of 'set'
        """
        self.conn.update_stage.side_effect = ClientError(error_content, "update_stage")
        result = boto_apigateway.activate_api_deployment(
            restApiId="rm06h9oac4",
            stageName="test",
            deploymentId="n05smo",
            **conn_parameters
        )
        self.assertFalse(result.get("set"))

    def test_that_when_creating_an_api_deployment_succeeds_the_create_api_deployment_method_returns_true(
        self,
    ):
        """
        tests that we can successfully create an api deployment and the createDate is
        converted to string
        """
        now = datetime.datetime.now()
        self.conn.create_deployment.return_value = {
            "description": "test-lambda-api-key",
            "id": "n05smo",
            "createdDate": now,
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }

        result = boto_apigateway.create_api_deployment(
            restApiId="rm06h9oac4", stageName="test", **conn_parameters
        )
        deployment = result.get("deployment")
        now_str = "{}".format(now)

        self.assertTrue(result.get("created"))
        self.assertEqual(deployment.get("createdDate"), now_str)

    def test_that_when_creating_an_deployment_fails_the_create_api_deployment_method_returns_error(
        self,
    ):
        """
        tests that we properly handle errors when create an api deployment fails.
        """

        self.conn.create_deployment.side_effect = ClientError(
            error_content, "create_deployment"
        )
        result = boto_apigateway.create_api_deployment(
            restApiId="rm06h9oac4", stageName="test", **conn_parameters
        )
        self.assertIs(result.get("created"), False)
        self.assertEqual(
            result.get("error").get("message"),
            error_message.format("create_deployment"),
        )

    def test_that_when_deleting_an_api_deployment_that_exists_the_delete_api_deployment_method_returns_true(
        self,
    ):
        """
        test True if the api deployment is successfully deleted.
        """
        self.conn.delete_deployment.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            }
        }
        result = boto_apigateway.delete_api_deployment(
            restApiId="rm06h9oac4", deploymentId="n05smo", **conn_parameters
        )
        self.assertTrue(result.get("deleted"))

    def test_that_when_deleting_an_api_deployment_that_does_not_exist_the_delete_api_deployment_method_returns_false(
        self,
    ):
        """
        Test that the given api deployment doesn't exists, and delete_api_deployment should return deleted status of False
        """
        self.conn.delete_deployment.side_effect = ClientError(
            error_content, "delete_deployment"
        )
        result = boto_apigateway.delete_api_deployment(
            restApiId="rm06h9oac4", deploymentId="n05smo1", **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_describing_api_stages_the_describe_api_stages_method_returns_list_of_stages(
        self,
    ):
        """
        Test Equality for number of stages for the given deployment is 2
        """
        self.conn.get_stages.return_value = {
            "item": [
                {
                    "cacheClusterEnabled": False,
                    "cacheClusterStatus": "NOT_AVAILABLE",
                    "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
                    "deploymentId": "n05smo",
                    "description": "test",
                    "lastUpdatedDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
                    "stageName": "test",
                },
                {
                    "cacheClusterEnabled": False,
                    "cacheClusterStatus": "NOT_AVAILABLE",
                    "createdDate": datetime.datetime(2015, 12, 17, 16, 33, 50),
                    "deploymentId": "n05smo",
                    "description": "dev",
                    "lastUpdatedDate": datetime.datetime(2015, 12, 17, 16, 33, 50),
                    "stageName": "dev",
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        result = boto_apigateway.describe_api_stages(
            restApiId="rm06h9oac4", deploymentId="n05smo", **conn_parameters
        )
        self.assertEqual(len(result.get("stages", {})), 2)

    def test_that_when_describing_api_stages_and_that_the_deployment_does_not_exist_the_describe_api_stages_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_stages.side_effect = ClientError(error_content, "get_stages")
        result = boto_apigateway.describe_api_stages(
            restApiId="rm06h9oac4", deploymentId="n05smo", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_stages")
        )

    def test_that_when_describing_an_api_stage_the_describe_api_stage_method_returns_the_stage(
        self,
    ):
        """
        Test True for the returned stage
        """
        self.conn.get_stage.return_value = {
            "cacheClusterEnabled": False,
            "cacheClusterStatus": "NOT_AVAILABLE",
            "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "deploymentId": "n05smo",
            "description": "test",
            "lastUpdatedDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "stageName": "test",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        result = boto_apigateway.describe_api_stage(
            restApiId="rm06h9oac4", stageName="test", **conn_parameters
        )
        self.assertTrue(result.get("stage"))

    def test_that_when_describing_api_stage_that_does_not_exist_the_describe_api_stage_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_stage.side_effect = ClientError(error_content, "get_stage")
        result = boto_apigateway.describe_api_stage(
            restApiId="rm06h9oac4", stageName="no_such_stage", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_stage")
        )

    def test_that_when_overwriting_stage_variables_to_an_existing_stage_the_overwrite_api_stage_variables_method_returns_the_updated_stage(
        self,
    ):
        """
        Test True for the returned stage
        """
        self.conn.get_stage.return_value = {
            "cacheClusterEnabled": False,
            "cacheClusterStatus": "NOT_AVAILABLE",
            "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "deploymentId": "n05smo",
            "description": "test",
            "lastUpdatedDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "stageName": "test",
            "variables": {"key1": "val1"},
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        self.conn.update_stage.return_value = {
            "cacheClusterEnabled": False,
            "cacheClusterStatus": "NOT_AVAILABLE",
            "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "deploymentId": "n05smo",
            "description": "test",
            "lastUpdatedDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "stageName": "test",
            "variables": {"key1": "val2"},
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        result = boto_apigateway.overwrite_api_stage_variables(
            restApiId="rm06h9oac4",
            stageName="test",
            variables=dict(key1="val2"),
            **conn_parameters
        )
        self.assertEqual(result.get("stage").get("variables").get("key1"), "val2")

    def test_that_when_overwriting_stage_variables_to_a_nonexisting_stage_the_overwrite_api_stage_variables_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_stage.side_effect = ClientError(error_content, "get_stage")
        result = boto_apigateway.overwrite_api_stage_variables(
            restApiId="rm06h9oac4",
            stageName="no_such_stage",
            variables=dict(key1="val1", key2="val2"),
            **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_stage")
        )

    def test_that_when_overwriting_stage_variables_to_an_existing_stage_the_overwrite_api_stage_variables_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned due to update_stage
        """
        self.conn.get_stage.return_value = {
            "cacheClusterEnabled": False,
            "cacheClusterStatus": "NOT_AVAILABLE",
            "createdDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "deploymentId": "n05smo",
            "description": "test",
            "lastUpdatedDate": datetime.datetime(2015, 11, 17, 16, 33, 50),
            "stageName": "test",
            "variables": {"key1": "val1"},
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        self.conn.update_stage.side_effect = ClientError(error_content, "update_stage")
        result = boto_apigateway.overwrite_api_stage_variables(
            restApiId="rm06h9oac4",
            stageName="test",
            variables=dict(key1="val2"),
            **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("update_stage")
        )

    def test_that_when_creating_an_api_stage_succeeds_the_create_api_stage_method_returns_true(
        self,
    ):
        """
        tests that we can successfully create an api stage and the createDate is
        converted to string
        """
        now = datetime.datetime.now()
        self.conn.create_stage.return_value = {
            "cacheClusterEnabled": False,
            "cacheClusterStatus": "NOT_AVAILABLE",
            "createdDate": now,
            "deploymentId": "n05smo",
            "description": "test",
            "lastUpdatedDate": now,
            "stageName": "test",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }

        result = boto_apigateway.create_api_stage(
            restApiId="rm06h9oac4",
            stageName="test",
            deploymentId="n05smo",
            **conn_parameters
        )
        stage = result.get("stage")
        now_str = "{}".format(now)
        self.assertIs(result.get("created"), True)
        self.assertEqual(stage.get("createdDate"), now_str)
        self.assertEqual(stage.get("lastUpdatedDate"), now_str)

    def test_that_when_creating_an_api_stage_fails_the_create_api_stage_method_returns_error(
        self,
    ):
        """
        tests that we properly handle errors when create an api stage fails.
        """

        self.conn.create_stage.side_effect = ClientError(error_content, "create_stage")
        result = boto_apigateway.create_api_stage(
            restApiId="rm06h9oac4",
            stageName="test",
            deploymentId="n05smo",
            **conn_parameters
        )
        self.assertIs(result.get("created"), False)
        self.assertEqual(
            result.get("error").get("message"), error_message.format("create_stage")
        )

    def test_that_when_deleting_an_api_stage_that_exists_the_delete_api_stage_method_returns_true(
        self,
    ):
        """
        test True if the api stage is successfully deleted.
        """
        self.conn.delete_stage.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            }
        }
        result = boto_apigateway.delete_api_stage(
            restApiId="rm06h9oac4", stageName="test", **conn_parameters
        )
        self.assertTrue(result.get("deleted"))

    def test_that_when_deleting_an_api_stage_that_does_not_exist_the_delete_api_stage_method_returns_false(
        self,
    ):
        """
        Test that the given api stage doesn't exists, and delete_api_stage should return deleted status of False
        """
        self.conn.delete_stage.side_effect = ClientError(error_content, "delete_stage")
        result = boto_apigateway.delete_api_stage(
            restApiId="rm06h9oac4", stageName="no_such_stage", **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_flushing_api_stage_cache_for_an_existing_stage_the_flush_api_stage_cache_method_returns_true(
        self,
    ):
        """
        Test True for 'flushed'
        """
        self.conn.flush_stage_cache.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            }
        }
        result = boto_apigateway.flush_api_stage_cache(
            restApiId="rm06h9oac4", stageName="no_such_stage", **conn_parameters
        )
        self.assertTrue(result.get("flushed"))

    def test_that_when_flushing_api_stage_cache_and_the_stage_does_not_exist_the_flush_api_stage_cache_method_returns_false(
        self,
    ):
        """
        Test False for 'flushed'
        """
        self.conn.flush_stage_cache.side_effect = ClientError(
            error_content, "flush_stage_cache"
        )
        result = boto_apigateway.flush_api_stage_cache(
            restApiId="rm06h9oac4", stageName="no_such_stage", **conn_parameters
        )
        self.assertFalse(result.get("flushed"))

    def test_that_when_describing_api_models_the_describe_api_models_method_returns_list_of_models(
        self,
    ):
        """
        Test Equality for number of models for the given api is 2
        """
        self.conn.get_models.return_value = {
            "items": [
                {
                    "contentType": "application/json",
                    "name": "Error",
                    "description": "Error Model",
                    "id": "iltqcb",
                    "schema": '{"properties":{"code":{"type":"integer","format":"int32"},"message":{"type":"string"},"fields":{"type":"string"}},"definitions":{}}',
                },
                {
                    "contentType": "application/json",
                    "name": "User",
                    "description": "User Model",
                    "id": "iltqcc",
                    "schema": (
                        '{"properties":{"username":{"type":"string","description":"A'
                        " unique username for the"
                        ' user"},"password":{"type":"string","description":"A password'
                        ' for the new user"}},"definitions":{}}'
                    ),
                },
            ],
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            },
        }
        result = boto_apigateway.describe_api_models(
            restApiId="rm06h9oac4", **conn_parameters
        )
        self.assertEqual(len(result.get("models", {})), 2)

    def test_that_when_describing_api_models_and_that_the_api_does_not_exist_the_describe_api_models_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_models.side_effect = ClientError(error_content, "get_models")
        result = boto_apigateway.describe_api_models(
            restApiId="rm06h9oac4", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_models")
        )

    def test_that_when_describing_api_model_the_describe_api_model_method_returns_the_model(
        self,
    ):
        """
        Test True for the returned stage
        """
        self.conn.get_model.return_value = api_model_ret
        result = boto_apigateway.describe_api_model(
            restApiId="rm06h9oac4", modelName="Error", **conn_parameters
        )
        self.assertTrue(result.get("model"))

    def test_that_when_describing_api_model_and_that_the_model_does_not_exist_the_describe_api_model_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_model.side_effect = ClientError(error_content, "get_model")
        result = boto_apigateway.describe_api_model(
            restApiId="rm06h9oac4", modelName="Error", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_model")
        )

    def test_that_model_exists_the_api_model_exists_method_returns_true(self):
        """
        Tests True when model exists
        """
        self.conn.get_model.return_value = api_model_ret
        result = boto_apigateway.api_model_exists(
            restApiId="rm06h9oac4", modelName="Error", **conn_parameters
        )
        self.assertTrue(result.get("exists"))

    def test_that_model_does_not_exists_the_api_model_exists_method_returns_false(self):
        """
        Tests False when model does not exist
        """
        self.conn.get_model.side_effect = ClientError(error_content, "get_model")
        result = boto_apigateway.api_model_exists(
            restApiId="rm06h9oac4", modelName="Error", **conn_parameters
        )
        self.assertFalse(result.get("exists"))

    def test_that_updating_model_schema_the_update_api_model_schema_method_returns_true(
        self,
    ):
        """
        Tests True when model schema is updated.
        """
        self.conn.update_model.return_value = api_model_ret
        result = boto_apigateway.update_api_model_schema(
            restApiId="rm06h9oac4",
            modelName="Error",
            schema=api_model_error_schema,
            **conn_parameters
        )
        self.assertTrue(result.get("updated"))

    def test_that_updating_model_schema_when_model_does_not_exist_the_update_api_model_schema_emthod_returns_false(
        self,
    ):
        """
        Tests False when model schema is not upated.
        """
        self.conn.update_model.side_effect = ClientError(error_content, "update_model")
        result = boto_apigateway.update_api_model_schema(
            restApiId="rm06h9oac4",
            modelName="no_such_model",
            schema=api_model_error_schema,
            **conn_parameters
        )
        self.assertFalse(result.get("updated"))

    def test_that_when_creating_an_api_model_succeeds_the_create_api_model_method_returns_true(
        self,
    ):
        """
        tests that we can successfully create an api model
        """
        self.conn.create_model.return_value = api_model_ret
        result = boto_apigateway.create_api_model(
            restApiId="rm06h9oac4",
            modelName="Error",
            modelDescription="Error Model",
            schema=api_model_error_schema,
            **conn_parameters
        )
        self.assertTrue(result.get("created"))

    def test_that_when_creating_an_api_model_fails_the_create_api_model_method_returns_error(
        self,
    ):
        """
        tests that we properly handle errors when create an api model fails.
        """
        self.conn.create_model.side_effect = ClientError(error_content, "create_model")
        result = boto_apigateway.create_api_model(
            restApiId="rm06h9oac4",
            modelName="Error",
            modelDescription="Error Model",
            schema=api_model_error_schema,
            **conn_parameters
        )
        self.assertFalse(result.get("created"))

    def test_that_when_deleting_an_api_model_that_exists_the_delete_api_model_method_returns_true(
        self,
    ):
        """
        test True if the api model is successfully deleted.
        """
        self.conn.delete_model.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            }
        }
        result = boto_apigateway.delete_api_model(
            restApiId="rm06h9oac4", modelName="Error", **conn_parameters
        )
        self.assertTrue(result.get("deleted"))

    def test_that_when_deleting_an_api_model_that_does_not_exist_the_delete_api_model_method_returns_false(
        self,
    ):
        """
        Test that the given api model doesn't exists, and delete_api_model should return deleted status of False
        """
        self.conn.delete_model.side_effect = ClientError(error_content, "delete_model")
        result = boto_apigateway.delete_api_model(
            restApiId="rm06h9oac4", modelName="no_such_model", **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_describing_api_resources_the_describe_api_resources_method_returns_list_of_3_resources(
        self,
    ):
        """
        Test Equality for number of resources for the given api is 3
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.describe_api_resources(
            restApiId="rm06h9oac4", **conn_parameters
        )
        self.assertEqual(
            len(result.get("resources")), len(api_resources_ret.get("items"))
        )

    def test_that_when_describing_api_resources_and_that_the_api_does_not_exist_the_describe_api_resources_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_resources.side_effect = ClientError(
            error_content, "get_resources"
        )
        result = boto_apigateway.describe_api_resources(
            restApiId="rm06h9oac4", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_resources")
        )

    def test_that_when_describing_an_api_resource_that_exists_the_describe_api_resource_method_returns_the_resource(
        self,
    ):
        """
        Test Equality of the resource path returned is /api
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.describe_api_resource(
            restApiId="rm06h9oac4", path="/api", **conn_parameters
        )
        self.assertEqual(result.get("resource", {}).get("path"), "/api")

    def test_that_when_describing_an_api_resource_that_does_not_exist_the_describe_api_resource_method_returns_the_resource_as_none(
        self,
    ):
        """
        Test Equality of the 'resource' is None
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.describe_api_resource(
            restApiId="rm06h9oac4", path="/path/does/not/exist", **conn_parameters
        )
        self.assertEqual(result.get("resource"), None)

    def test_that_when_describing_an_api_resource_and_that_the_api_does_not_exist_the_describe_api_resource_method_returns_error(
        self,
    ):
        """
        Test Equality of error returned
        """
        self.conn.get_resources.side_effect = ClientError(
            error_content, "get_resources"
        )
        result = boto_apigateway.describe_api_resource(
            restApiId="bad_id", path="/api", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_resources")
        )

    def test_that_when_creating_api_resources_for_a_path_that_creates_one_new_resource_the_create_resources_api_method_returns_all_resources(
        self,
    ):
        """
        Tests that a path of '/api3' returns 2 resources, named '/' and '/api'.
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.create_resource.return_value = api_create_resource_ret

        result = boto_apigateway.create_api_resources(
            restApiId="rm06h9oac4", path="/api3", **conn_parameters
        )

        resources = result.get("resources")
        self.assertIs(result.get("created"), True)
        self.assertEqual(len(resources), 2)
        self.assertEqual(resources[0].get("path"), "/")
        self.assertEqual(resources[1].get("path"), "/api3")

    def test_that_when_creating_api_resources_for_a_path_whose_resources_exist_the_create_resources_api_method_returns_all_resources(
        self,
    ):
        """
        Tests that a path of '/api/users' as defined in api_resources_ret return resources named '/', '/api',
        and '/api/users'
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.create_api_resources(
            restApiId="rm06h9oac4", path="/api/users", **conn_parameters
        )
        resources = result.get("resources")
        self.assertIs(result.get("created"), True)
        self.assertEqual(len(resources), len(api_resources_ret.get("items")))
        self.assertEqual(resources[0].get("path"), "/")
        self.assertEqual(resources[1].get("path"), "/api")
        self.assertEqual(resources[2].get("path"), "/api/users")

    def test_that_when_creating_api_resource_fails_the_create_resources_api_method_returns_false(
        self,
    ):
        """
        Tests False if we failed to create a resource
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.create_resource.side_effect = ClientError(
            error_content, "create_resource"
        )
        result = boto_apigateway.create_api_resources(
            restApiId="rm06h9oac4", path="/api4", **conn_parameters
        )
        self.assertFalse(result.get("created"))

    def test_that_when_deleting_api_resources_for_a_resource_that_exists_the_delete_api_resources_method_returns_true(
        self,
    ):
        """
        Tests True for '/api'
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.delete_api_resources(
            restApiId="rm06h9oac4", path="/api", **conn_parameters
        )
        self.assertTrue(result.get("deleted"))

    def test_that_when_deleting_api_resources_for_a_resource_that_does_not_exist_the_delete_api_resources_method_returns_false(
        self,
    ):
        """
        Tests False for '/api5'
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.delete_api_resources(
            restApiId="rm06h9oac4", path="/api5", **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_deleting_the_root_api_resource_the_delete_api_resources_method_returns_false(
        self,
    ):
        """
        Tests False for '/'
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.delete_api_resources(
            restApiId="rm06h9oac4", path="/", **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_deleting_api_resources_and_delete_resource_throws_error_the_delete_api_resources_method_returns_false(
        self,
    ):
        """
        Tests False delete_resource side side_effect
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.delete_resource.side_effect = ClientError(
            error_content, "delete_resource"
        )
        result = boto_apigateway.delete_api_resources(
            restApiId="rm06h9oac4", path="/api", **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_describing_an_api_resource_method_that_exists_the_describe_api_resource_method_returns_the_method(
        self,
    ):
        """
        Tests True for '/api/users' and POST
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_method.return_value = {
            "httpMethod": "POST",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }
        result = boto_apigateway.describe_api_resource_method(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            **conn_parameters
        )
        self.assertTrue(result.get("method"))

    def test_that_when_describing_an_api_resource_method_whose_method_does_not_exist_the_describe_api_resource_method_returns_error(
        self,
    ):
        """
        Tests Equality of returned error for '/api/users' and PUT
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_method.side_effect = ClientError(error_content, "get_method")
        result = boto_apigateway.describe_api_resource_method(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="PUT",
            **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_method")
        )

    def test_that_when_describing_an_api_resource_method_whose_resource_does_not_exist_the_describe_api_resrouce_method_returns_error(
        self,
    ):
        """
        Tests True for resource not found error for '/does/not/exist' and POST
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.describe_api_resource_method(
            restApiId="rm06h9oac4",
            resourcePath="/does/not/exist",
            httpMethod="POST",
            **conn_parameters
        )
        self.assertTrue(result.get("error"))

    def test_that_when_creating_an_api_method_the_create_api_method_method_returns_true(
        self,
    ):
        """
        Tests True on 'created' for '/api/users' and 'GET'
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.put_method.return_value = {
            "httpMethod": "GET",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }
        result = boto_apigateway.create_api_method(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="GET",
            authorizationType="NONE",
            **conn_parameters
        )
        self.assertTrue(result.get("created"))

    def test_that_when_creating_an_api_method_and_resource_does_not_exist_the_create_api_method_method_returns_false(
        self,
    ):
        """
        Tests False on 'created' for '/api5', and 'GET'
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.create_api_method(
            restApiId="rm06h9oac4",
            resourcePath="/api5",
            httpMethod="GET",
            authorizationType="NONE",
            **conn_parameters
        )
        self.assertFalse(result.get("created"))

    def test_that_when_creating_an_api_method_and_error_thrown_on_put_method_the_create_api_method_method_returns_false(
        self,
    ):
        """
        Tests False on 'created' for '/api/users' and 'GET'
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.put_method.side_effect = ClientError(error_content, "put_method")
        result = boto_apigateway.create_api_method(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="GET",
            authorizationType="NONE",
            **conn_parameters
        )
        self.assertFalse(result.get("created"))

    def test_that_when_deleting_an_api_method_for_a_method_that_exist_the_delete_api_method_method_returns_true(
        self,
    ):
        """
        Tests True for '/api/users' and 'POST'
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.delete_method.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            }
        }
        result = boto_apigateway.delete_api_method(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            **conn_parameters
        )
        self.assertTrue(result.get("deleted"))

    def test_that_when_deleting_an_api_method_for_a_method_that_does_not_exist_the_delete_api_method_method_returns_false(
        self,
    ):
        """
        Tests False for '/api/users' and 'GET'
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.delete_method.side_effect = ClientError(
            error_content, "delete_method"
        )
        result = boto_apigateway.delete_api_method(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="GET",
            **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_deleting_an_api_method_for_a_resource_that_does_not_exist_the_delete_api_method_method_returns_false(
        self,
    ):
        """
        Tests False for '/api/users5' and 'POST'
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.delete_api_method(
            restApiId="rm06h9oac4",
            resourcePath="/api/users5",
            httpMethod="POST",
            **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_describing_an_api_method_response_that_exists_the_describe_api_method_respond_method_returns_the_response(
        self,
    ):
        """
        Tests True for 'response' for '/api/users', 'POST', and 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_method_response.return_value = {
            "statusCode": 200,
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }
        result = boto_apigateway.describe_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            statusCode=200,
            **conn_parameters
        )
        self.assertTrue(result.get("response"))

    def test_that_when_describing_an_api_method_response_and_response_code_does_not_exist_the_describe_api_method_response_method_returns_error(
        self,
    ):
        """
        Tests Equality of error msg thrown from get_method_response for '/api/users', 'POST', and 250
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_method_response.side_effect = ClientError(
            error_content, "get_method_response"
        )
        result = boto_apigateway.describe_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            statusCode=250,
            **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"),
            error_message.format("get_method_response"),
        )

    def test_that_when_describing_an_api_method_response_and_resource_does_not_exist_the_describe_api_method_response_method_returns_error(
        self,
    ):
        """
        Tests True for existence of 'error' for '/api5/users', 'POST', and 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.describe_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api5/users",
            httpMethod="POST",
            statusCode=200,
            **conn_parameters
        )
        self.assertTrue(result.get("error"))

    def test_that_when_creating_an_api_method_response_the_create_api_method_response_method_returns_true(
        self,
    ):
        """
        Tests True on 'created' for '/api/users', 'POST', 201
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.put_method_response.return_value = {
            "statusCode": "201",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }
        result = boto_apigateway.create_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            statusCode="201",
            **conn_parameters
        )
        self.assertTrue(result.get("created"))

    def test_that_when_creating_an_api_method_response_and_resource_does_not_exist_the_create_api_method_response_method_returns_false(
        self,
    ):
        """
        Tests False on 'created' for '/api5', 'POST', 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.create_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api5",
            httpMethod="POST",
            statusCode="200",
            **conn_parameters
        )
        self.assertFalse(result.get("created"))

    def test_that_when_creating_an_api_method_response_and_error_thrown_on_put_method_response_the_create_api_method_response_method_returns_false(
        self,
    ):
        """
        Tests False on 'created' for '/api/users', 'POST', 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.put_method_response.side_effect = ClientError(
            error_content, "put_method_response"
        )
        result = boto_apigateway.create_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            statusCode="200",
            **conn_parameters
        )
        self.assertFalse(result.get("created"))

    def test_that_when_deleting_an_api_method_response_for_a_response_that_exist_the_delete_api_method_response_method_returns_true(
        self,
    ):
        """
        Tests True for '/api/users', 'POST', 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.delete_method_response.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "2d31072c-9d15-11e5-9977-6d9fcfda9c0a",
            }
        }
        result = boto_apigateway.delete_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            statusCode="200",
            **conn_parameters
        )
        self.assertTrue(result.get("deleted"))

    def test_that_when_deleting_an_api_method_response_for_a_response_that_does_not_exist_the_delete_api_method_response_method_returns_false(
        self,
    ):
        """
        Tests False for '/api/users', 'POST', 201
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.delete_method_response.side_effect = ClientError(
            error_content, "delete_method_response"
        )
        result = boto_apigateway.delete_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="GET",
            statusCode="201",
            **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_deleting_an_api_method_response_for_a_resource_that_does_not_exist_the_delete_api_method_response_method_returns_false(
        self,
    ):
        """
        Tests False for '/api/users5', 'POST', 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.delete_api_method_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users5",
            httpMethod="POST",
            statusCode="200",
            **conn_parameters
        )
        self.assertFalse(result.get("deleted"))

    def test_that_when_describing_an_api_integration_that_exists_the_describe_api_integration_method_returns_the_intgration(
        self,
    ):
        """
        Tests True for 'integration' for '/api/users', 'POST'
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_integration.return_value = {
            "type": "AWS",
            "uri": "arn:aws:apigateway:us-west-2:lambda:path/2015-03-31/functions/arn:aws:lambda:us-west-2:1234568992820:function:echo_event/invocations",
            "credentials": "testing",
            "httpMethod": "POST",
            "intgrationResponses": {"200": {}},
            "requestTemplates": {"application/json": {}},
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }
        result = boto_apigateway.describe_api_integration(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            **conn_parameters
        )
        self.assertTrue(result.get("integration"))

    def test_that_when_describing_an_api_integration_and_method_does_not_have_integration_defined_the_describe_api_integration_method_returns_error(
        self,
    ):
        """
        Tests Equality of error msg thrown from get_method_response for '/api/users', 'GET'
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_integration.side_effect = ClientError(
            error_content, "get_integration"
        )
        result = boto_apigateway.describe_api_integration(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="GET",
            **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format("get_integration")
        )

    def test_that_when_describing_an_api_integration_and_resource_does_not_exist_the_describe_api_integration_method_returns_error(
        self,
    ):
        """
        Tests True for existence of 'error' for '/api5/users', 'POST'
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.describe_api_integration(
            restApiId="rm06h9oac4",
            resourcePath="/api5/users",
            httpMethod="POST",
            **conn_parameters
        )
        self.assertTrue(result.get("error"))

    def test_that_when_describing_an_api_integration_response_that_exists_the_describe_api_integration_response_method_returns_the_intgration(
        self,
    ):
        """
        Tests True for 'response' for '/api/users', 'POST', 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_integration_response.return_value = {
            "responseParameters": {},
            "statusCode": 200,
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
                "RequestId": "7cc233dd-9dc8-11e5-ba47-1b7350cc2757",
            },
        }
        result = boto_apigateway.describe_api_integration_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            statusCode="200",
            **conn_parameters
        )
        self.assertTrue(result.get("response"))

    def test_that_when_describing_an_api_integration_response_and_status_code_does_not_exist_the_describe_api_integration_response_method_returns_error(
        self,
    ):
        """
        Tests Equality of error msg thrown from get_method_response for '/api/users', 'POST', 201
        """
        self.conn.get_resources.return_value = api_resources_ret
        self.conn.get_integration_response.side_effect = ClientError(
            error_content, "get_integration_response"
        )
        result = boto_apigateway.describe_api_integration_response(
            restApiId="rm06h9oac4",
            resourcePath="/api/users",
            httpMethod="POST",
            statusCode="201",
            **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"),
            error_message.format("get_integration_response"),
        )

    def test_that_when_describing_an_api_integration_response_and_resource_does_not_exist_the_describe_api_integration_response_method_returns_error(
        self,
    ):
        """
        Tests True for existence of 'error' for '/api5/users', 'POST', 200
        """
        self.conn.get_resources.return_value = api_resources_ret
        result = boto_apigateway.describe_api_integration_response(
            restApiId="rm06h9oac4",
            resourcePath="/api5/users",
            httpMethod="POST",
            statusCode="200",
            **conn_parameters
        )
        self.assertTrue(result.get("error"))

    def test_that_when_describing_usage_plans_and_an_exception_is_thrown_in_get_usage_plans(
        self,
    ):
        """
        Tests True for existence of 'error'
        """
        self.conn.get_usage_plans.side_effect = ClientError(
            error_content, "get_usage_plans_exception"
        )
        result = boto_apigateway.describe_usage_plans(
            name="some plan", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"),
            error_message.format("get_usage_plans_exception"),
        )

    def test_that_when_describing_usage_plans_and_plan_name_or_id_does_not_exist_that_results_have_empty_plans_list(
        self,
    ):
        """
        Tests for plans equaling empty list
        """
        self.conn.get_usage_plans.return_value = usage_plans_ret

        result = boto_apigateway.describe_usage_plans(
            name="does not exist", **conn_parameters
        )
        self.assertEqual(result.get("plans"), [])

        result = boto_apigateway.describe_usage_plans(
            plan_id="does not exist", **conn_parameters
        )
        self.assertEqual(result.get("plans"), [])

        result = boto_apigateway.describe_usage_plans(
            name="does not exist", plan_id="does not exist", **conn_parameters
        )
        self.assertEqual(result.get("plans"), [])

        result = boto_apigateway.describe_usage_plans(
            name="plan1_name", plan_id="does not exist", **conn_parameters
        )
        self.assertEqual(result.get("plans"), [])

        result = boto_apigateway.describe_usage_plans(
            name="does not exist", plan_id="plan1_id", **conn_parameters
        )
        self.assertEqual(result.get("plans"), [])

    def test_that_when_describing_usage_plans_for_plans_that_exist_that_the_function_returns_all_matching_plans(
        self,
    ):
        """
        Tests for plans filtering properly if they exist
        """
        self.conn.get_usage_plans.return_value = usage_plans_ret

        result = boto_apigateway.describe_usage_plans(
            name=usage_plan1["name"], **conn_parameters
        )
        self.assertEqual(len(result.get("plans")), 2)
        for plan in result["plans"]:
            self.assertTrue(plan in [usage_plan1, usage_plan1b])

    def test_that_when_creating_or_updating_a_usage_plan_and_throttle_or_quota_failed_to_validate_that_an_error_is_returned(
        self,
    ):
        """
        Tests for TypeError and ValueError in throttle and quota
        """
        for throttle, quota in (([], None), (None, []), ("abc", None), (None, "def")):
            res = boto_apigateway.create_usage_plan(
                "plan1_name",
                description=None,
                throttle=throttle,
                quota=quota,
                **conn_parameters
            )
            self.assertNotEqual(None, res.get("error"))
            res = boto_apigateway.update_usage_plan(
                "plan1_id", throttle=throttle, quota=quota, **conn_parameters
            )
            self.assertNotEqual(None, res.get("error"))

        for quota in ({"limit": 123}, {"period": 123}, {"period": "DAY"}):
            res = boto_apigateway.create_usage_plan(
                "plan1_name",
                description=None,
                throttle=None,
                quota=quota,
                **conn_parameters
            )
            self.assertNotEqual(None, res.get("error"))
            res = boto_apigateway.update_usage_plan(
                "plan1_id", quota=quota, **conn_parameters
            )
            self.assertNotEqual(None, res.get("error"))

        self.assertTrue(self.conn.get_usage_plans.call_count == 0)
        self.assertTrue(self.conn.create_usage_plan.call_count == 0)
        self.assertTrue(self.conn.update_usage_plan.call_count == 0)

    def test_that_when_creating_a_usage_plan_and_create_usage_plan_throws_an_exception_that_an_error_is_returned(
        self,
    ):
        """
        tests for ClientError
        """
        self.conn.create_usage_plan.side_effect = ClientError(
            error_content, "create_usage_plan_exception"
        )
        result = boto_apigateway.create_usage_plan(name="some plan", **conn_parameters)
        self.assertEqual(
            result.get("error").get("message"),
            error_message.format("create_usage_plan_exception"),
        )

    def test_that_create_usage_plan_succeeds(self):
        """
        tests for success user plan creation
        """
        res = "unit test create_usage_plan succeeded"
        self.conn.create_usage_plan.return_value = res
        result = boto_apigateway.create_usage_plan(name="some plan", **conn_parameters)
        self.assertEqual(result.get("created"), True)
        self.assertEqual(result.get("result"), res)

    def test_that_when_udpating_a_usage_plan_and_update_usage_plan_throws_an_exception_that_an_error_is_returned(
        self,
    ):
        """
        tests for ClientError
        """
        self.conn.update_usage_plan.side_effect = ClientError(
            error_content, "update_usage_plan_exception"
        )
        result = boto_apigateway.update_usage_plan(
            plan_id="plan1_id", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"),
            error_message.format("update_usage_plan_exception"),
        )

    def test_that_when_updating_a_usage_plan_and_if_throttle_and_quota_parameters_are_none_update_usage_plan_removes_throttle_and_quota(
        self,
    ):
        """
        tests for throttle and quota removal
        """
        ret = "some success status"
        self.conn.update_usage_plan.return_value = ret
        result = boto_apigateway.update_usage_plan(
            plan_id="plan1_id", throttle=None, quota=None, **conn_parameters
        )
        self.assertEqual(result.get("updated"), True)
        self.assertEqual(result.get("result"), ret)
        self.assertTrue(self.conn.update_usage_plan.call_count >= 1)

    def test_that_when_deleting_usage_plan_and_describe_usage_plans_had_error_that_the_same_error_is_returned(
        self,
    ):
        """
        tests for error in describe_usage_plans returns error
        """
        ret = "get_usage_plans_exception"
        self.conn.get_usage_plans.side_effect = ClientError(error_content, ret)
        result = boto_apigateway.delete_usage_plan(
            plan_id="some plan id", **conn_parameters
        )
        self.assertEqual(result.get("error").get("message"), error_message.format(ret))
        self.assertTrue(self.conn.delete_usage_plan.call_count == 0)

    def test_that_when_deleting_usage_plan_and_plan_exists_that_the_functions_returns_deleted_true(
        self,
    ):
        self.conn.get_usage_plans.return_value = usage_plans_ret
        ret = "delete_usage_plan_retval"
        self.conn.delete_usage_plan.return_value = ret
        result = boto_apigateway.delete_usage_plan(
            plan_id="plan1_id", **conn_parameters
        )
        self.assertEqual(result.get("deleted"), True)
        self.assertEqual(result.get("usagePlanId"), "plan1_id")
        self.assertTrue(self.conn.delete_usage_plan.call_count >= 1)

    def test_that_when_deleting_usage_plan_and_plan_does_not_exist_that_the_functions_returns_deleted_true(
        self,
    ):
        """
        tests for ClientError
        """
        self.conn.get_usage_plans.return_value = dict(items=[])
        ret = "delete_usage_plan_retval"
        self.conn.delete_usage_plan.return_value = ret
        result = boto_apigateway.delete_usage_plan(
            plan_id="plan1_id", **conn_parameters
        )
        self.assertEqual(result.get("deleted"), True)
        self.assertEqual(result.get("usagePlanId"), "plan1_id")
        self.assertTrue(self.conn.delete_usage_plan.call_count == 0)

    def test_that_when_deleting_usage_plan_and_delete_usage_plan_throws_exception_that_an_error_is_returned(
        self,
    ):
        """
        tests for ClientError
        """
        self.conn.get_usage_plans.return_value = usage_plans_ret
        error_msg = "delete_usage_plan_exception"
        self.conn.delete_usage_plan.side_effect = ClientError(error_content, error_msg)
        result = boto_apigateway.delete_usage_plan(
            plan_id="plan1_id", **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format(error_msg)
        )
        self.assertTrue(self.conn.delete_usage_plan.call_count >= 1)

    def test_that_attach_or_detach_usage_plan_when_apis_is_empty_that_success_is_returned(
        self,
    ):
        """
        tests for border cases when apis is empty list
        """
        result = boto_apigateway.attach_usage_plan_to_apis(
            plan_id="plan1_id", apis=[], **conn_parameters
        )
        self.assertEqual(result.get("success"), True)
        self.assertEqual(result.get("result", "no result?"), None)
        self.assertTrue(self.conn.update_usage_plan.call_count == 0)

        result = boto_apigateway.detach_usage_plan_from_apis(
            plan_id="plan1_id", apis=[], **conn_parameters
        )
        self.assertEqual(result.get("success"), True)
        self.assertEqual(result.get("result", "no result?"), None)
        self.assertTrue(self.conn.update_usage_plan.call_count == 0)

    def test_that_attach_or_detach_usage_plan_when_api_does_not_contain_apiId_or_stage_that_an_error_is_returned(
        self,
    ):
        """
        tests for invalid key in api object
        """
        for api in ({"apiId": "some Id"}, {"stage": "some stage"}, {}):
            result = boto_apigateway.attach_usage_plan_to_apis(
                plan_id="plan1_id", apis=[api], **conn_parameters
            )
            self.assertNotEqual(result.get("error"), None)

            result = boto_apigateway.detach_usage_plan_from_apis(
                plan_id="plan1_id", apis=[api], **conn_parameters
            )
            self.assertNotEqual(result.get("error"), None)

        self.assertTrue(self.conn.update_usage_plan.call_count == 0)

    def test_that_attach_or_detach_usage_plan_and_update_usage_plan_throws_exception_that_an_error_is_returned(
        self,
    ):
        """
        tests for ClientError
        """
        api = {"apiId": "some_id", "stage": "some_stage"}
        error_msg = "update_usage_plan_exception"
        self.conn.update_usage_plan.side_effect = ClientError(error_content, error_msg)

        result = boto_apigateway.attach_usage_plan_to_apis(
            plan_id="plan1_id", apis=[api], **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format(error_msg)
        )

        result = boto_apigateway.detach_usage_plan_from_apis(
            plan_id="plan1_id", apis=[api], **conn_parameters
        )
        self.assertEqual(
            result.get("error").get("message"), error_message.format(error_msg)
        )

    def test_that_attach_or_detach_usage_plan_updated_successfully(self):
        """
        tests for update_usage_plan called
        """
        api = {"apiId": "some_id", "stage": "some_stage"}
        attach_ret = "update_usage_plan_add_op_succeeded"
        detach_ret = "update_usage_plan_remove_op_succeeded"
        self.conn.update_usage_plan.side_effect = [attach_ret, detach_ret]

        result = boto_apigateway.attach_usage_plan_to_apis(
            plan_id="plan1_id", apis=[api], **conn_parameters
        )
        self.assertEqual(result.get("success"), True)
        self.assertEqual(result.get("result"), attach_ret)

        result = boto_apigateway.detach_usage_plan_from_apis(
            plan_id="plan1_id", apis=[api], **conn_parameters
        )
        self.assertEqual(result.get("success"), True)
        self.assertEqual(result.get("result"), detach_ret)



import salt.modules.nexus as nexus
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class nexusTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {nexus: {}}

    def test_artifact_get_metadata(self):
        with patch(
            "salt.modules.nexus._get_artifact_metadata_xml",
            MagicMock(
                return_value="""<?xml version="1.0" encoding="UTF-8"?>
<metadata>
  <groupId>com.company.sampleapp.web-module</groupId>
  <artifactId>web</artifactId>
  <versioning>
    <release>0.1.0</release>
    <versions>
      <version>0.0.1</version>
      <version>0.0.2</version>
      <version>0.0.3</version>
      <version>0.1.0</version>
    </versions>
    <lastUpdated>20171010143552</lastUpdated>
  </versioning>
</metadata>"""
            ),
        ):
            metadata = nexus._get_artifact_metadata(
                nexus_url="http://nexus.example.com/repository",
                repository="libs-releases",
                group_id="com.company.sampleapp.web-module",
                artifact_id="web",
                headers={},
            )
            self.assertEqual(metadata["latest_version"], "0.1.0")

    def test_snapshot_version_get_metadata(self):
        with patch(
            "salt.modules.nexus._get_snapshot_version_metadata_xml",
            MagicMock(
                return_value="""<?xml version="1.0" encoding="UTF-8"?>
<metadata modelVersion="1.1.0">
  <groupId>com.company.sampleapp.web-module</groupId>
  <artifactId>web</artifactId>
  <version>0.0.2-SNAPSHOT</version>
  <versioning>
    <snapshot>
      <timestamp>20170920.212353</timestamp>
      <buildNumber>3</buildNumber>
    </snapshot>
    <lastUpdated>20171112171500</lastUpdated>
    <snapshotVersions>
      <snapshotVersion>
        <classifier>sans-externalized</classifier>
        <extension>jar</extension>
        <value>0.0.2-20170920.212353-3</value>
        <updated>20170920212353</updated>
      </snapshotVersion>
      <snapshotVersion>
        <classifier>dist</classifier>
        <extension>zip</extension>
        <value>0.0.2-20170920.212353-3</value>
        <updated>20170920212353</updated>
      </snapshotVersion>
    </snapshotVersions>
  </versioning>
</metadata>"""
            ),
        ):
            metadata = nexus._get_snapshot_version_metadata(
                nexus_url="http://nexus.example.com/repository",
                repository="libs-releases",
                group_id="com.company.sampleapp.web-module",
                artifact_id="web",
                version="0.0.2-SNAPSHOT",
                headers={},
            )
            self.assertEqual(
                metadata["snapshot_versions"]["dist"], "0.0.2-20170920.212353-3"
            )

    def test_artifact_metadata_url(self):
        metadata_url = nexus._get_artifact_metadata_url(
            nexus_url="http://nexus.example.com/repository",
            repository="libs-releases",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
        )

        self.assertEqual(
            metadata_url,
            "http://nexus.example.com/repository/libs-releases/com/company/sampleapp/web-module/web/maven-metadata.xml",
        )

    def test_snapshot_version_metadata_url(self):
        metadata_url = nexus._get_snapshot_version_metadata_url(
            nexus_url="http://nexus.example.com/repository",
            repository="libs-snapshots",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            version="0.0.2-SNAPSHOT",
        )

        self.assertEqual(
            metadata_url,
            "http://nexus.example.com/repository/libs-snapshots/com/company/sampleapp/web-module/web/0.0.2-SNAPSHOT/maven-metadata.xml",
        )

    def test_construct_url_for_released_version(self):
        artifact_url, file_name = nexus._get_release_url(
            repository="libs-releases",
            group_id="com.company.sampleapp.web-module",
            artifact_id="web",
            packaging="zip",
            nexus_url="http://nexus.example.com/repository",
            version="0.1.0",
        )

        self.assertEqual(
            artifact_url,
            "http://nexus.example.com/repository/libs-releases/com/company/sampleapp/web-module/web/0.1.0/web-0.1.0.zip",
        )
        self.assertEqual(file_name, "web-0.1.0.zip")

    def test_construct_url_for_snapshot_version(self):
        with patch(
            "salt.modules.nexus._get_snapshot_version_metadata",
            MagicMock(
                return_value={"snapshot_versions": {"zip": "0.0.2-20170920.212353-3"}}
            ),
        ):

            artifact_url, file_name = nexus._get_snapshot_url(
                nexus_url="http://nexus.example.com/repository",
                repository="libs-snapshots",
                group_id="com.company.sampleapp.web-module",
                artifact_id="web",
                version="0.2.0-SNAPSHOT",
                packaging="zip",
                headers={},
            )

            self.assertEqual(
                artifact_url,
                "http://nexus.example.com/repository/libs-snapshots/com/company/sampleapp/web-module/web/0.2.0-SNAPSHOT/web-0.0.2-20170920.212353-3.zip",
            )
            self.assertEqual(file_name, "web-0.0.2-20170920.212353-3.zip")



"""
Module used to access the vcenter proxy connection methods
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)

__proxyenabled__ = ["vcenter"]
# Define the module's virtual name
__virtualname__ = "vcenter"


def __virtual__():
    """
    Only work on proxy
    """
    if salt.utils.platform.is_proxy():
        return __virtualname__
    return False


def get_details():
    return __proxy__["vcenter.get_details"]()



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import types
from datetime import datetime

import salt.modules.win_system as win_system
import salt.utils.platform
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase, skipIf

try:
    import wmi

    HAS_WMI = True
except ImportError:
    HAS_WMI = False


class MockWMI_ComputerSystem:
    """
    Mock WMI Win32_ComputerSystem Class
    """

    BootupState = "Normal boot"
    Caption = "SALT SERVER"
    ChassisBootupState = 3
    ChassisSKUNumber = "3.14159"
    DNSHostname = "SALT SERVER"
    Domain = "WORKGROUP"
    DomainRole = 2
    Manufacturer = "Dell Inc."
    Model = "Dell 2980"
    NetworkServerModeEnabled = True
    PartOfDomain = False
    PCSystemType = 4
    PowerState = 0
    Status = "OK"
    SystemType = "x64-based PC"
    TotalPhysicalMemory = 17078214656
    ThermalState = 3
    Workgroup = "WORKGROUP"

    def __init__(self):
        pass

    @staticmethod
    def Rename(Name):
        return Name == Name

    @staticmethod
    def JoinDomainOrWorkgroup(Name):
        return [0]

    @staticmethod
    def UnjoinDomainOrWorkgroup(Password, UserName, FUnjoinOptions):
        return [0]


class MockWMI_OperatingSystem:
    """
    Mock WMI Win32_OperatingSystem Class
    """

    Description = "Because salt goes EVERYWHERE"
    InstallDate = "20110211131800"
    LastBootUpTime = "19620612120000"
    Manufacturer = "Python"
    Caption = "Salty"
    NumberOfUsers = 7530000000
    Organization = "SaltStack"
    OSArchitecture = "Windows"
    Primary = True
    ProductType = 3
    RegisteredUser = "thatch@saltstack.com"
    SystemDirectory = "C:\\Windows\\System32"
    SystemDrive = "C:\\"
    Version = "10.0.17763"
    WindowsDirectory = "C:\\Windows"

    def __init__(self):
        pass


class MockWMI_ComputerSystemProduct:
    def __init__(self):
        self.SKUNumber = None


class MockWMI_Processor:
    """
    Mock WMI Win32_Processor Class
    """

    Manufacturer = "Intel"
    MaxClockSpeed = 2301
    NumberOfLogicalProcessors = 8
    NumberOfCores = 4
    NumberOfEnabledCore = 4

    def __init__(self):
        pass


class MockWMI_BIOS:
    """
    Mock WMI Win32_BIOS Class
    """

    SerialNumber = "SALTY2011"
    Manufacturer = "Dell Inc."
    Version = "DELL - 10283849"
    Caption = "A12"
    BIOSVersion = [Version, Caption, "ASUS - 3948D"]
    Description = Caption

    def __init__(self):
        pass


@skipIf(not HAS_WMI, "WMI only available on Windows")
@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinSystemTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_system
    """

    def setup_loader_modules(self):
        modules_globals = {}
        # wmi and pythoncom modules are platform specific...
        mock_pythoncom = types.ModuleType(salt.utils.stringutils.to_str("pythoncom"))
        sys_modules_patcher = patch.dict("sys.modules", {"pythoncom": mock_pythoncom})
        sys_modules_patcher.start()
        self.addCleanup(sys_modules_patcher.stop)
        self.WMI = Mock()
        self.addCleanup(delattr, self, "WMI")
        modules_globals["wmi"] = wmi

        if win_system.HAS_WIN32NET_MODS is False:
            win32api = types.ModuleType("win32api")
            now = datetime.now()
            win32api.GetLocalTime = MagicMock(
                return_value=[
                    now.year,
                    now.month,
                    now.weekday(),
                    now.day,
                    now.hour,
                    now.minute,
                    now.second,
                    now.microsecond,
                ]
            )
            modules_globals["win32api"] = win32api
            win32net = types.ModuleType("win32net")
            win32net.NetServerGetInfo = MagicMock()
            win32net.NetServerSetInfo = MagicMock()
            modules_globals["win32net"] = win32net

        return {win_system: modules_globals}

    def test_halt(self):
        """
        Test to halt a running system
        """
        mock = MagicMock(return_value="salt")
        with patch.object(win_system, "shutdown", mock):
            self.assertEqual(win_system.halt(), "salt")

    def test_init(self):
        """
        Test to change the system runlevel on sysV compatible systems
        """
        self.assertEqual(win_system.init(3), "Not implemented on Windows at this time.")

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_poweroff(self):
        """
        Test to poweroff a running system
        """
        mock = MagicMock(return_value="salt")
        with patch.object(win_system, "shutdown", mock):
            self.assertEqual(win_system.poweroff(), "salt")

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_reboot(self):
        """
        Test to reboot the system
        """
        with patch("salt.modules.win_system.shutdown", MagicMock(return_value=True)):
            self.assertEqual(win_system.reboot(), True)

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_reboot_with_timeout_in_minutes(self):
        """
        Test to reboot the system with a timeout
        """
        with patch(
            "salt.modules.win_system.shutdown", MagicMock(return_value=True)
        ) as shutdown:
            self.assertEqual(win_system.reboot(5, in_seconds=False), True)
            shutdown.assert_called_with(
                timeout=5, in_seconds=False, reboot=True, only_on_pending_reboot=False
            )

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_reboot_with_timeout_in_seconds(self):
        """
        Test to reboot the system with a timeout
        """
        with patch(
            "salt.modules.win_system.shutdown", MagicMock(return_value=True)
        ) as shutdown:
            self.assertEqual(win_system.reboot(5, in_seconds=True), True)
            shutdown.assert_called_with(
                timeout=5, in_seconds=True, reboot=True, only_on_pending_reboot=False
            )

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_reboot_with_wait(self):
        """
        Test to reboot the system with a timeout and
        wait for it to finish
        """
        with patch(
            "salt.modules.win_system.shutdown", MagicMock(return_value=True)
        ), patch("salt.modules.win_system.time.sleep", MagicMock()) as time:
            self.assertEqual(win_system.reboot(wait_for_reboot=True), True)
            time.assert_called_with(330)

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_shutdown(self):
        """
        Test to shutdown a running system
        """
        with patch(
            "salt.modules.win_system.win32api.InitiateSystemShutdown", MagicMock()
        ):
            self.assertEqual(win_system.shutdown(), True)

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_shutdown_hard(self):
        """
        Test to shutdown a running system with no timeout or warning
        """
        with patch(
            "salt.modules.win_system.shutdown", MagicMock(return_value=True)
        ) as shutdown:
            self.assertEqual(win_system.shutdown_hard(), True)
            shutdown.assert_called_with(timeout=0)

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_set_computer_name(self):
        """
        Test to set the Windows computer name
        """
        with patch(
            "salt.modules.win_system.windll.kernel32.SetComputerNameExW",
            MagicMock(return_value=True),
        ), patch.object(
            win_system, "get_computer_name", MagicMock(return_value="salt")
        ), patch.object(
            win_system,
            "get_pending_computer_name",
            MagicMock(return_value="salt_new"),
        ):
            self.assertDictEqual(
                win_system.set_computer_name("salt_new"),
                {"Computer Name": {"Current": "salt", "Pending": "salt_new"}},
            )
        # Test set_computer_name failure
        with patch(
            "salt.modules.win_system.windll.kernel32.SetComputerNameExW",
            MagicMock(return_value=False),
        ):
            self.assertFalse(win_system.set_computer_name("salt"))

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_set_computer_desc(self):
        """
        Test to set the Windows computer description
        """
        mock = MagicMock()
        mock_get_info = MagicMock(return_value={"comment": ""})
        mock_get_desc = MagicMock(return_value="Salt's comp")
        with patch(
            "salt.modules.win_system.win32net.NetServerGetInfo", mock_get_info
        ), patch(
            "salt.modules.win_system.win32net.NetServerSetInfo", mock
        ), patch.object(
            win_system, "get_computer_desc", mock_get_desc
        ):
            self.assertDictEqual(
                win_system.set_computer_desc("Salt's comp"),
                {"Computer Description": "Salt's comp"},
            )

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_get_computer_desc(self):
        """
        Test to get the Windows computer description
        """
        with patch(
            "salt.modules.win_system.get_system_info",
            MagicMock(
                side_effect=[{"description": "salt description"}, {"description": None}]
            ),
        ):
            self.assertEqual(win_system.get_computer_desc(), "salt description")
            self.assertFalse(win_system.get_computer_desc())

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_join_domain(self):
        """
        Test to join a computer to an Active Directory domain
        """
        with patch(
            "salt.modules.win_system._join_domain", MagicMock(return_value=0)
        ), patch(
            "salt.modules.win_system.get_domain_workgroup",
            MagicMock(return_value={"Workgroup": "Workgroup"}),
        ):
            self.assertDictEqual(
                win_system.join_domain("saltstack", "salt", "salt@123"),
                {"Domain": "saltstack", "Restart": False},
            )

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_join_domain_already_joined(self):
        """
        Test to join a computer to an Active Directory domain when it is
        already joined
        """
        with patch(
            "salt.modules.win_system._join_domain", MagicMock(return_value=0)
        ), patch(
            "salt.modules.win_system.get_domain_workgroup",
            MagicMock(return_value={"Domain": "saltstack"}),
        ):
            self.assertEqual(
                win_system.join_domain("saltstack", "salt", "salt@123"),
                "Already joined to saltstack",
            )

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_unjoin_domain(self):
        """
        Test unjoining a computer from an Active Directory domain
        """
        with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
            self.WMI, "Win32_ComputerSystem", return_value=[MockWMI_ComputerSystem()]
        ), patch.object(wmi, "WMI", Mock(return_value=self.WMI)), patch(
            "salt.modules.win_system.get_domain_workgroup",
            MagicMock(return_value={"Domain": "contoso.com"}),
        ):
            self.assertDictEqual(
                win_system.unjoin_domain(),
                {"Workgroup": "WORKGROUP", "Restart": False},
            )

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_unjoin_domain_already_unjoined(self):
        """
        Test unjoining a computer from an Active Directory domain
        """
        with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
            self.WMI, "Win32_ComputerSystem", return_value=[MockWMI_ComputerSystem()]
        ), patch.object(wmi, "WMI", Mock(return_value=self.WMI)), patch(
            "salt.modules.win_system.get_domain_workgroup",
            MagicMock(return_value={"Workgroup": "WORKGROUP"}),
        ):
            self.assertEqual(win_system.unjoin_domain(), "Already joined to WORKGROUP")

    def test_get_system_time(self):
        """
        Test to get system time
        """
        tm = datetime.strftime(datetime.now(), "%I:%M:%S %p")
        win_tm = win_system.get_system_time()
        try:
            self.assertEqual(win_tm, tm)
        except AssertionError:
            # handle race condition
            import re

            self.assertTrue(re.search(r"^\d{2}:\d{2} \w{2}$", win_tm))

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_set_system_time(self):
        """
        Test to set system time
        """
        with patch(
            "salt.modules.win_system.set_system_date_time",
            MagicMock(side_effect=[False, True]),
        ):
            self.assertFalse(win_system.set_system_time("11:31:15 AM"))
            self.assertTrue(win_system.set_system_time("11:31:15 AM"))

    def test_get_system_date(self):
        """
        Test to get system date
        """
        date = datetime.strftime(datetime.now(), "%m/%d/%Y")
        self.assertEqual(win_system.get_system_date(), date)

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_set_system_date(self):
        """
        Test to set system date
        """
        with patch(
            "salt.modules.win_system.set_system_date_time",
            MagicMock(side_effect=[False, True]),
        ):
            self.assertFalse(win_system.set_system_date("03-28-13"))
            self.assertTrue(win_system.set_system_date("03-28-13"))

    def test_start_time_service(self):
        """
        Test to start the Windows time service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(win_system.__salt__, {"service.start": mock}):
            self.assertTrue(win_system.start_time_service())

    def test_stop_time_service(self):
        """
        Test to stop the windows time service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(win_system.__salt__, {"service.stop": mock}):
            self.assertTrue(win_system.stop_time_service())

    def test_set_hostname(self):
        """
        Test setting a new hostname
        """
        with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
            self.WMI, "Win32_ComputerSystem", return_value=[MockWMI_ComputerSystem()]
        ), patch.object(wmi, "WMI", Mock(return_value=self.WMI)):
            self.assertTrue(win_system.set_hostname("NEW"))

    def test_get_domain_workgroup(self):
        """
        Test get_domain_workgroup
        """
        with patch.object(wmi, "WMI", Mock(return_value=self.WMI)), patch(
            "salt.utils.winapi.Com", MagicMock()
        ), patch.object(
            self.WMI, "Win32_ComputerSystem", return_value=[MockWMI_ComputerSystem()]
        ):
            self.assertDictEqual(
                win_system.get_domain_workgroup(), {"Workgroup": "WORKGROUP"}
            )

    def test_set_domain_workgroup(self):
        """
        Test set_domain_workgroup
        """
        with patch.object(wmi, "WMI", Mock(return_value=self.WMI)), patch(
            "salt.utils.winapi.Com", MagicMock()
        ), patch.object(
            self.WMI, "Win32_ComputerSystem", return_value=[MockWMI_ComputerSystem()]
        ):
            self.assertTrue(win_system.set_domain_workgroup("test"))

    def test_get_hostname(self):
        """
        Test getting a new hostname
        """
        cmd_run_mock = MagicMock(return_value="MINION")
        with patch.dict(win_system.__salt__, {"cmd.run": cmd_run_mock}):
            ret = win_system.get_hostname()
            self.assertEqual(ret, "MINION")
        cmd_run_mock.assert_called_once_with(cmd="hostname")

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_get_system_info(self):
        fields = [
            "bios_caption",
            "bios_description",
            "bios_details",
            "bios_manufacturer",
            "bios_version",
            "bootup_state",
            "caption",
            "chassis_bootup_state",
            "chassis_sku_number",
            "description",
            "dns_hostname",
            "domain",
            "domain_role",
            "hardware_manufacturer",
            "hardware_model",
            "hardware_serial",
            "install_date",
            "last_boot",
            "name",
            "network_server_mode_enabled",
            "organization",
            "os_architecture",
            "os_manufacturer",
            "os_name",
            "os_type",
            "os_version",
            "part_of_domain",
            "pc_system_type",
            "power_state",
            "primary",
            "processor_cores",
            "processor_cores_enabled",
            "processor_manufacturer",
            "processor_max_clock_speed",
            "processors",
            "processors_logical",
            "registered_user",
            "status",
            "system_directory",
            "system_drive",
            "system_type",
            "thermal_state",
            "total_physical_memory",
            "total_physical_memory_raw",
            "users",
            "windows_directory",
            "workgroup",
        ]
        with patch("salt.utils.win_system.get_computer_name", MagicMock()), patch(
            "salt.utils.winapi.Com", MagicMock()
        ), patch.object(
            self.WMI, "Win32_OperatingSystem", return_value=[MockWMI_OperatingSystem()]
        ), patch.object(
            self.WMI, "Win32_ComputerSystem", return_value=[MockWMI_ComputerSystem()]
        ), patch.object(
            self.WMI,
            "Win32_ComputerSystemProduct",
            return_value=[MockWMI_ComputerSystemProduct()],
        ), patch.object(
            self.WMI,
            "Win32_Processor",
            return_value=[MockWMI_Processor(), MockWMI_Processor()],
        ), patch.object(
            self.WMI, "Win32_BIOS", return_value=[MockWMI_BIOS()]
        ), patch.object(
            wmi, "WMI", Mock(return_value=self.WMI)
        ):
            ret = win_system.get_system_info()
        # Make sure all the fields are in the return
        for field in fields:
            self.assertIn(field, ret)
        # os_type
        os_types = ["Work Station", "Domain Controller", "Server"]
        self.assertIn(ret["os_type"], os_types)
        domain_roles = [
            "Standalone Workstation",
            "Member Workstation",
            "Standalone Server",
            "Member Server",
            "Backup Domain Controller",
            "Primary Domain Controller",
        ]
        self.assertIn(ret["domain_role"], domain_roles)
        system_types = [
            "Unspecified",
            "Desktop",
            "Mobile",
            "Workstation",
            "Enterprise Server",
            "SOHO Server",
            "Appliance PC",
            "Performance Server",
            "Slate",
            "Maximum",
        ]
        self.assertIn(ret["pc_system_type"], system_types)
        warning_states = [
            "Other",
            "Unknown",
            "Safe",
            "Warning",
            "Critical",
            "Non-recoverable",
        ]
        self.assertIn(ret["chassis_bootup_state"], warning_states)
        self.assertIn(ret["thermal_state"], warning_states)

    @skipIf(not win_system.HAS_WIN32NET_MODS, "Missing win32 libraries")
    def test_get_system_info_no_number_of_enabled_core(self):
        """
        Tests get_system_info when there is no `NumberOfEnabledCore` property in
        the WMI Class Win32_Processor. Older versions of Windows are missing
        this property
        """
        # Create a mock processor class that does not have the
        # NumberOfCoresEnabled property
        class MockWMIProcessor:
            """
            Mock WMI Win32_Processor Class
            """

            def __init__(self):
                self.Manufacturer = "Intel"
                self.MaxClockSpeed = 2301
                self.NumberOfLogicalProcessors = 8
                self.NumberOfCores = 4

        with patch("salt.utils.win_system.get_computer_name", MagicMock()), patch(
            "salt.utils.winapi.Com", MagicMock()
        ), patch.object(
            self.WMI, "Win32_OperatingSystem", return_value=[MockWMI_OperatingSystem()]
        ), patch.object(
            self.WMI, "Win32_ComputerSystem", return_value=[MockWMI_ComputerSystem()]
        ), patch.object(
            self.WMI,
            "Win32_ComputerSystemProduct",
            return_value=[MockWMI_ComputerSystemProduct()],
        ), patch.object(
            self.WMI,
            "Win32_Processor",
            return_value=[MockWMIProcessor(), MockWMIProcessor()],
        ), patch.object(
            self.WMI, "Win32_BIOS", return_value=[MockWMI_BIOS()]
        ), patch.object(
            wmi, "WMI", Mock(return_value=self.WMI)
        ):
            ret = win_system.get_system_info()
            self.assertIn("processors", ret)
            self.assertIn("processors_logical", ret)
            self.assertIn("processor_cores", ret)
            self.assertIn("processor_manufacturer", ret)
            self.assertIn("processor_max_clock_speed", ret)
            self.assertNotIn("processor_cores_enabled", ret)



"""
Manage DACLs on Windows

:depends:   - winreg Python module
"""


import logging
import os
import re

import salt.utils.platform
from salt.exceptions import CommandExecutionError

# TODO: Figure out the exceptions that could be raised and properly catch
#       them instead of a bare except that catches any exception at all
#       may also need to add the ability to take ownership of an object to set
#       permissions if the minion is running as a user and not LOCALSYSTEM


try:
    import winreg
    import win32security
    import ntsecuritycon

    HAS_WINDOWS_MODULES = True
except ImportError:
    HAS_WINDOWS_MODULES = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "win_dacl"


class daclConstants:
    """
    DACL constants used throughout the module
    """

    # Definition in ntsecuritycon is incorrect (does not match winnt.h). The version
    # in ntsecuritycon has the extra bits 0x200 enabled.
    # Note that you when you set this permission what you'll generally get back is it
    # ORed with 0x200 (SI_NO_ACL_PROTECT), which is what ntsecuritycon incorrectly defines.

    def __init__(self):
        self.FILE_ALL_ACCESS = (
            ntsecuritycon.STANDARD_RIGHTS_REQUIRED | ntsecuritycon.SYNCHRONIZE | 0x1FF
        )

        self.hkeys_security = {
            "HKEY_LOCAL_MACHINE": "MACHINE",
            "HKEY_USERS": "USERS",
            "HKEY_CURRENT_USER": "CURRENT_USER",
            "HKEY_CLASSES_ROOT": "CLASSES_ROOT",
            "MACHINE": "MACHINE",
            "USERS": "USERS",
            "CURRENT_USER": "CURRENT_USER",
            "CLASSES_ROOT": "CLASSES_ROOT",
            "HKLM": "MACHINE",
            "HKU": "USERS",
            "HKCU": "CURRENT_USER",
            "HKCR": "CLASSES_ROOT",
        }
        self.rights = {
            win32security.SE_REGISTRY_KEY: {
                "READ": {"BITS": winreg.KEY_READ, "TEXT": "read"},
                "FULLCONTROL": {"BITS": winreg.KEY_ALL_ACCESS, "TEXT": "full control"},
            },
            win32security.SE_FILE_OBJECT: {
                "READ": {"BITS": ntsecuritycon.FILE_GENERIC_READ, "TEXT": "read"},
                "WRITE": {"BITS": ntsecuritycon.FILE_GENERIC_WRITE, "TEXT": "write"},
                "READ&EXECUTE": {
                    "BITS": ntsecuritycon.FILE_GENERIC_EXECUTE
                    | ntsecuritycon.FILE_GENERIC_READ,
                    "TEXT": "read and execute",
                },
                "MODIFY": {
                    "BITS": ntsecuritycon.FILE_GENERIC_WRITE
                    | ntsecuritycon.FILE_GENERIC_READ
                    | ntsecuritycon.FILE_GENERIC_EXECUTE
                    | ntsecuritycon.DELETE,
                    "TEXT": "modify",
                },
                "FULLCONTROL": {"BITS": self.FILE_ALL_ACCESS, "TEXT": "full control"},
            },
        }
        self.validAceTypes = {
            "ALLOW": {"TEXT": "allowed", "BITS": 0},
            "DENY": {"TEXT": "denied", "BITS": 1},
        }
        self.validPropagations = {
            win32security.SE_REGISTRY_KEY: {
                "KEY": {"TEXT": "this key only", "BITS": win32security.NO_INHERITANCE},
                "KEY&SUBKEYS": {
                    "TEXT": "this key and subkeys",
                    "BITS": win32security.CONTAINER_INHERIT_ACE,
                },
                "SUBKEYS": {
                    "TEXT": "subkeys only",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.CONTAINER_INHERIT_ACE,
                },
                "THIS KEY ONLY": {
                    "TEXT": "this key only",
                    "BITS": win32security.NO_INHERITANCE,
                },
                "THIS KEY AND SUBKEYS": {
                    "TEXT": "this key and subkeys",
                    "BITS": win32security.CONTAINER_INHERIT_ACE,
                },
                "SUBKEYS ONLY": {
                    "TEXT": "subkeys only",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.CONTAINER_INHERIT_ACE,
                },
            },
            win32security.SE_FILE_OBJECT: {
                "FILE": {
                    "TEXT": "this file/folder only",
                    "BITS": win32security.NO_INHERITANCE,
                },
                "FOLDER": {
                    "TEXT": "this file/folder only",
                    "BITS": win32security.NO_INHERITANCE,
                },
                "FOLDER&SUBFOLDERS&FILES": {
                    "TEXT": "this folder, subfolders, and files",
                    "BITS": win32security.CONTAINER_INHERIT_ACE
                    | win32security.OBJECT_INHERIT_ACE,
                },
                "FOLDER&SUBFOLDERS": {
                    "TEXT": "this folder and subfolders",
                    "BITS": win32security.CONTAINER_INHERIT_ACE,
                },
                "FOLDER&FILES": {
                    "TEXT": "this folder and files",
                    "BITS": win32security.OBJECT_INHERIT_ACE,
                },
                "SUBFOLDERS&FILES": {
                    "TEXT": "subfolders and files",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.CONTAINER_INHERIT_ACE
                    | win32security.OBJECT_INHERIT_ACE,
                },
                "SUBFOLDERS": {
                    "TEXT": "subfolders only",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.CONTAINER_INHERIT_ACE,
                },
                "FILES": {
                    "TEXT": "files only",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.OBJECT_INHERIT_ACE,
                },
                "THIS FILE ONLY": {
                    "TEXT": "this file/folder only",
                    "BITS": win32security.NO_INHERITANCE,
                },
                "THIS FOLDER ONLY": {
                    "TEXT": "this file/folder only",
                    "BITS": win32security.NO_INHERITANCE,
                },
                "THIS FOLDER, SUBFOLDERS, AND FILES": {
                    "TEXT": "this folder, subfolders, and files",
                    "BITS": win32security.CONTAINER_INHERIT_ACE
                    | win32security.OBJECT_INHERIT_ACE,
                },
                "THIS FOLDER AND SUBFOLDERS": {
                    "TEXT": "this folder and subfolders",
                    "BITS": win32security.CONTAINER_INHERIT_ACE,
                },
                "THIS FOLDER AND FILES": {
                    "TEXT": "this folder and files",
                    "BITS": win32security.OBJECT_INHERIT_ACE,
                },
                "SUBFOLDERS AND FILES": {
                    "TEXT": "subfolders and files",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.CONTAINER_INHERIT_ACE
                    | win32security.OBJECT_INHERIT_ACE,
                },
                "SUBFOLDERS ONLY": {
                    "TEXT": "subfolders only",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.CONTAINER_INHERIT_ACE,
                },
                "FILES ONLY": {
                    "TEXT": "files only",
                    "BITS": win32security.INHERIT_ONLY_ACE
                    | win32security.OBJECT_INHERIT_ACE,
                },
            },
        }
        self.reflection_mask = {
            True: winreg.KEY_ALL_ACCESS,
            False: winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY,
        }
        self.objectType = {
            "FILE": win32security.SE_FILE_OBJECT,
            "DIRECTORY": win32security.SE_FILE_OBJECT,
            "REGISTRY": win32security.SE_REGISTRY_KEY,
        }

    def getObjectTypeBit(self, t):
        """
        returns the bit value of the string object type
        """
        if isinstance(t, str):
            t = t.upper()
            try:
                return self.objectType[t]
            except KeyError:
                raise CommandExecutionError(
                    'Invalid object type "{}". It should be one of the following:  {}'.format(
                        t, ", ".join(self.objectType)
                    )
                )
        else:
            return t

    def getSecurityHkey(self, s):
        """
        returns the necessary string value for an HKEY for the win32security module
        """
        try:
            return self.hkeys_security[s]
        except KeyError:
            raise CommandExecutionError(
                'No HKEY named "{}".  It should be one of the following:  {}'.format(
                    s, ", ".join(self.hkeys_security)
                )
            )

    def getPermissionBit(self, t, m):
        """
        returns a permission bit of the string permission value for the specified object type
        """
        try:
            if isinstance(m, str):
                return self.rights[t][m]["BITS"]
            else:
                return m
        except KeyError:
            raise CommandExecutionError(
                'No right "{}".  It should be one of the following:  {}'.format(
                    m, ", ".join(self.rights[t])
                )
            )

    def getPermissionText(self, t, m):
        """
        returns the permission textual representation of a specified permission bit/object type
        """
        try:
            return self.rights[t][m]["TEXT"]
        except KeyError:
            raise CommandExecutionError(
                'No right "{}".  It should be one of the following:  {}'.format(
                    m, ", ".join(self.rights[t])
                )
            )

    def getAceTypeBit(self, t):
        """
        returns the acetype bit of a text value
        """
        try:
            return self.validAceTypes[t]["BITS"]
        except KeyError:
            raise CommandExecutionError(
                'No ACE type "{}".  It should be one of the following:  {}'.format(
                    t, ", ".join(self.validAceTypes)
                )
            )

    def getAceTypeText(self, t):
        """
        returns the textual representation of a acetype bit
        """
        try:
            return self.validAceTypes[t]["TEXT"]
        except KeyError:
            raise CommandExecutionError(
                'No ACE type "{}".  It should be one of the following:  {}'.format(
                    t, ", ".join(self.validAceTypes)
                )
            )

    def getPropagationBit(self, t, p):
        """
        returns the propagation bit of a text value
        """
        try:
            return self.validPropagations[t][p]["BITS"]
        except KeyError:
            raise CommandExecutionError(
                'No propagation type of "{}".  It should be one of the following:  {}'.format(
                    p, ", ".join(self.validPropagations[t])
                )
            )

    def getPropagationText(self, t, p):
        """
        returns the textual representation of a propagation bit
        """
        try:
            return self.validPropagations[t][p]["TEXT"]
        except KeyError:
            raise CommandExecutionError(
                'No propagation type of "{}".  It should be one of the following:  {}'.format(
                    p, ", ".join(self.validPropagations[t])
                )
            )

    def processPath(self, path, objectType):
        """
        processes a path/object type combo and returns:
            registry types with the correct HKEY text representation
            files/directories with environment variables expanded
        """
        if objectType == win32security.SE_REGISTRY_KEY:
            splt = path.split("\\")
            hive = self.getSecurityHkey(splt.pop(0).upper())
            splt.insert(0, hive)
            path = r"\\".join(splt)
        else:
            path = os.path.expandvars(path)
        return path


def _getUserSid(user):
    """
    return a state error dictionary, with 'sid' as a field if it could be returned
    if user is None, sid will also be None
    """
    ret = {}

    sid_pattern = r"^S-1(-\d+){1,}$"

    if user and re.match(sid_pattern, user, re.I):
        try:
            sid = win32security.GetBinarySid(user)
        except Exception as e:  # pylint: disable=broad-except
            ret["result"] = False
            ret["comment"] = (
                "Unable to obtain the binary security identifier for {}.  The exception"
                " was {}.".format(user, e)
            )
        else:
            try:
                win32security.LookupAccountSid("", sid)
                ret["result"] = True
                ret["sid"] = sid
            except Exception as e:  # pylint: disable=broad-except
                ret["result"] = False
                ret["comment"] = (
                    "Unable to lookup the account for the security identifier {}.  The"
                    " exception was {}.".format(user, e)
                )
    else:
        try:
            sid = win32security.LookupAccountName("", user)[0] if user else None
            ret["result"] = True
            ret["sid"] = sid
        except Exception as e:  # pylint: disable=broad-except
            ret["result"] = False
            ret["comment"] = (
                "Unable to obtain the security identifier for {}.  The exception"
                " was {}.".format(user, e)
            )
    return ret


def __virtual__():
    """
    Only works on Windows systems
    """
    if salt.utils.platform.is_windows() and HAS_WINDOWS_MODULES:
        return __virtualname__
    return (False, "Module win_dacl: module only works on Windows systems")


def _get_dacl(path, objectType):
    """
    Gets the DACL of a path
    """
    try:
        dacl = win32security.GetNamedSecurityInfo(
            path, objectType, win32security.DACL_SECURITY_INFORMATION
        ).GetSecurityDescriptorDacl()
    except Exception:  # pylint: disable=broad-except
        dacl = None
    return dacl


def get(path, objectType, user=None):
    """
    Get the ACL of an object. Will filter by user if one is provided.

    Args:
        path: The path to the object
        objectType: The type of object (FILE, DIRECTORY, REGISTRY)
        user: A user name to filter by

    Returns (dict): A dictionary containing the ACL

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' win_dacl.get c:\temp directory
    """
    ret = {"Path": path, "ACLs": []}

    sidRet = _getUserSid(user)

    if path and objectType:
        dc = daclConstants()
        objectTypeBit = dc.getObjectTypeBit(objectType)
        path = dc.processPath(path, objectTypeBit)
        tdacl = _get_dacl(path, objectTypeBit)
        if tdacl:
            for counter in range(0, tdacl.GetAceCount()):
                tAce = tdacl.GetAce(counter)
                if not sidRet["sid"] or (tAce[2] == sidRet["sid"]):
                    ret["ACLs"].append(_ace_to_text(tAce, objectTypeBit))
    return ret


def add_ace(path, objectType, user, permission, acetype, propagation):
    r"""
    add an ace to an object

    path:  path to the object (i.e. c:\\temp\\file, HKEY_LOCAL_MACHINE\\SOFTWARE\\KEY, etc)
    user: user to add
    permission:  permissions for the user
    acetype:  either allow/deny for each user/permission (ALLOW, DENY)
    propagation: how the ACE applies to children for Registry Keys and Directories(KEY, KEY&SUBKEYS, SUBKEYS)

    CLI Example:

    .. code-block:: bash

        allow domain\fakeuser full control on HKLM\\SOFTWARE\\somekey, propagate to this key and subkeys
            salt 'myminion' win_dacl.add_ace 'HKEY_LOCAL_MACHINE\\SOFTWARE\\somekey' 'Registry' 'domain\fakeuser' 'FULLCONTROL' 'ALLOW' 'KEY&SUBKEYS'
    """
    ret = {"result": None, "changes": {}, "comment": ""}

    if path and user and permission and acetype and propagation:
        if objectType.upper() == "FILE":
            propagation = "FILE"
        dc = daclConstants()
        objectTypeBit = dc.getObjectTypeBit(objectType)
        path = dc.processPath(path, objectTypeBit)
        user = user.strip()
        permission = permission.strip().upper()
        acetype = acetype.strip().upper()
        propagation = propagation.strip().upper()

        sidRet = _getUserSid(user)
        if not sidRet["result"]:
            return sidRet
        permissionbit = dc.getPermissionBit(objectTypeBit, permission)
        acetypebit = dc.getAceTypeBit(acetype)
        propagationbit = dc.getPropagationBit(objectTypeBit, propagation)
        dacl = _get_dacl(path, objectTypeBit)

        if dacl:
            acesAdded = []
            try:
                if acetypebit == 0:
                    dacl.AddAccessAllowedAceEx(
                        win32security.ACL_REVISION,
                        propagationbit,
                        permissionbit,
                        sidRet["sid"],
                    )
                elif acetypebit == 1:
                    dacl.AddAccessDeniedAceEx(
                        win32security.ACL_REVISION,
                        propagationbit,
                        permissionbit,
                        sidRet["sid"],
                    )
                win32security.SetNamedSecurityInfo(
                    path,
                    objectTypeBit,
                    win32security.DACL_SECURITY_INFORMATION,
                    None,
                    None,
                    dacl,
                    None,
                )
                acesAdded.append(
                    "{} {} {} on {}".format(
                        user,
                        dc.getAceTypeText(acetype),
                        dc.getPermissionText(objectTypeBit, permission),
                        dc.getPropagationText(objectTypeBit, propagation),
                    )
                )
                ret["result"] = True
            except Exception as e:  # pylint: disable=broad-except
                ret[
                    "comment"
                ] = "An error occurred attempting to add the ace.  The error was {}".format(
                    e
                )
                ret["result"] = False
                return ret
            if acesAdded:
                ret["changes"]["Added ACEs"] = acesAdded
        else:
            ret["comment"] = "Unable to obtain the DACL of {}".format(path)
    else:
        ret["comment"] = "An empty value was specified for a required item."
        ret["result"] = False
    return ret


def rm_ace(path, objectType, user, permission=None, acetype=None, propagation=None):
    r"""
    remove an ace to an object

    path:  path to the object (i.e. c:\\temp\\file, HKEY_LOCAL_MACHINE\\SOFTWARE\\KEY, etc)
    user: user to remove
    permission:  permissions for the user
    acetypes:  either allow/deny for each user/permission (ALLOW, DENY)
    propagation: how the ACE applies to children for Registry Keys and Directories(KEY, KEY&SUBKEYS, SUBKEYS)

    If any of the optional parameters are omitted (or set to None) they act as wildcards.

    CLI Example:

    .. code-block:: bash

        remove allow domain\fakeuser full control on HKLM\\SOFTWARE\\somekey propagated to this key and subkeys
            salt 'myminion' win_dacl.rm_ace 'Registry' 'HKEY_LOCAL_MACHINE\\SOFTWARE\\somekey' 'domain\fakeuser' 'FULLCONTROL' 'ALLOW' 'KEY&SUBKEYS'
    """
    ret = {"result": None, "changes": {}, "comment": ""}

    if path and user:
        dc = daclConstants()
        if propagation and objectType.upper() == "FILE":
            propagation = "FILE"
        objectTypeBit = dc.getObjectTypeBit(objectType)
        path = dc.processPath(path, objectTypeBit)

        user = user.strip()
        permission = permission.strip().upper() if permission else None
        acetype = acetype.strip().upper() if acetype else None
        propagation = propagation.strip().upper() if propagation else None

        if check_ace(path, objectType, user, permission, acetype, propagation, True)[
            "Exists"
        ]:
            sidRet = _getUserSid(user)
            if not sidRet["result"]:
                return sidRet
            permissionbit = (
                dc.getPermissionBit(objectTypeBit, permission) if permission else None
            )
            acetypebit = dc.getAceTypeBit(acetype) if acetype else None
            propagationbit = (
                dc.getPropagationBit(objectTypeBit, propagation)
                if propagation
                else None
            )
            dacl = _get_dacl(path, objectTypeBit)
            counter = 0
            acesRemoved = []
            while counter < dacl.GetAceCount():
                tAce = dacl.GetAce(counter)
                if (
                    tAce[0][1] & win32security.INHERITED_ACE
                ) != win32security.INHERITED_ACE:
                    if tAce[2] == sidRet["sid"]:
                        if not acetypebit or tAce[0][0] == acetypebit:
                            if not propagationbit or (
                                (tAce[0][1] & propagationbit) == propagationbit
                            ):
                                if not permissionbit or tAce[1] == permissionbit:
                                    dacl.DeleteAce(counter)
                                    counter = counter - 1
                                    acesRemoved.append(
                                        _ace_to_text(tAce, objectTypeBit)
                                    )

                counter = counter + 1
            if acesRemoved:
                try:
                    win32security.SetNamedSecurityInfo(
                        path,
                        objectTypeBit,
                        win32security.DACL_SECURITY_INFORMATION,
                        None,
                        None,
                        dacl,
                        None,
                    )
                    ret["changes"]["Removed ACEs"] = acesRemoved
                    ret["result"] = True
                except Exception as e:  # pylint: disable=broad-except
                    ret["result"] = False
                    ret["comment"] = "Error removing ACE.  The error was {}.".format(e)
                    return ret
        else:
            ret["comment"] = "The specified ACE was not found on the path."
    return ret


def _ace_to_text(ace, objectType):
    """
    helper function to convert an ace to a textual representation
    """
    dc = daclConstants()
    objectType = dc.getObjectTypeBit(objectType)
    try:
        userSid = win32security.LookupAccountSid("", ace[2])
        if userSid[1]:
            userSid = "{1}\\{0}".format(userSid[0], userSid[1])
        else:
            userSid = "{}".format(userSid[0])
    except Exception:  # pylint: disable=broad-except
        userSid = win32security.ConvertSidToStringSid(ace[2])
    tPerm = ace[1]
    tAceType = ace[0][0]
    tProps = ace[0][1]
    tInherited = ""
    for x in dc.validAceTypes:
        if dc.validAceTypes[x]["BITS"] == tAceType:
            tAceType = dc.validAceTypes[x]["TEXT"]
            break
    for x in dc.rights[objectType]:
        if dc.rights[objectType][x]["BITS"] == tPerm:
            tPerm = dc.rights[objectType][x]["TEXT"]
            break
    if (tProps & win32security.INHERITED_ACE) == win32security.INHERITED_ACE:
        tInherited = "[Inherited]"
        tProps = tProps ^ win32security.INHERITED_ACE
    for x in dc.validPropagations[objectType]:
        if dc.validPropagations[objectType][x]["BITS"] == tProps:
            tProps = dc.validPropagations[objectType][x]["TEXT"]
            break
    return "{} {} {} on {} {}".format(userSid, tAceType, tPerm, tProps, tInherited)


def _set_dacl_inheritance(path, objectType, inheritance=True, copy=True, clear=False):
    """
    helper function to set the inheritance
    Args:

        path (str): The path to the object

        objectType (str): The type of object

        inheritance (bool): True enables inheritance, False disables

        copy (bool): Copy inherited ACEs to the DACL before disabling
        inheritance

        clear (bool): Remove non-inherited ACEs from the DACL
    """
    ret = {"result": False, "comment": "", "changes": {}}

    if path:
        try:
            sd = win32security.GetNamedSecurityInfo(
                path, objectType, win32security.DACL_SECURITY_INFORMATION
            )
            tdacl = sd.GetSecurityDescriptorDacl()
            if inheritance:
                if clear:
                    counter = 0
                    removedAces = []
                    while counter < tdacl.GetAceCount():
                        tAce = tdacl.GetAce(counter)
                        if (
                            tAce[0][1] & win32security.INHERITED_ACE
                        ) != win32security.INHERITED_ACE:
                            tdacl.DeleteAce(counter)
                            removedAces.append(_ace_to_text(tAce, objectType))
                        else:
                            counter = counter + 1
                    if removedAces:
                        ret["changes"]["Removed ACEs"] = removedAces
                else:
                    ret["changes"]["Non-Inherited ACEs"] = "Left in the DACL"
                win32security.SetNamedSecurityInfo(
                    path,
                    objectType,
                    win32security.DACL_SECURITY_INFORMATION
                    | win32security.UNPROTECTED_DACL_SECURITY_INFORMATION,
                    None,
                    None,
                    tdacl,
                    None,
                )
                ret["changes"]["Inheritance"] = "Enabled"
            else:
                if not copy:
                    counter = 0
                    inheritedAcesRemoved = []
                    while counter < tdacl.GetAceCount():
                        tAce = tdacl.GetAce(counter)
                        if (
                            tAce[0][1] & win32security.INHERITED_ACE
                        ) == win32security.INHERITED_ACE:
                            tdacl.DeleteAce(counter)
                            inheritedAcesRemoved.append(_ace_to_text(tAce, objectType))
                        else:
                            counter = counter + 1
                    if inheritedAcesRemoved:
                        ret["changes"]["Removed ACEs"] = inheritedAcesRemoved
                else:
                    ret["changes"]["Previously Inherited ACEs"] = "Copied to the DACL"
                win32security.SetNamedSecurityInfo(
                    path,
                    objectType,
                    win32security.DACL_SECURITY_INFORMATION
                    | win32security.PROTECTED_DACL_SECURITY_INFORMATION,
                    None,
                    None,
                    tdacl,
                    None,
                )
                ret["changes"]["Inheritance"] = "Disabled"
            ret["result"] = True
        except Exception as e:  # pylint: disable=broad-except
            ret["result"] = False
            ret[
                "comment"
            ] = "Error attempting to set the inheritance.  The error was {}.".format(e)

    return ret


def enable_inheritance(path, objectType, clear=False):
    """
    enable/disable inheritance on an object

    Args:
        path: The path to the object
        objectType: The type of object (FILE, DIRECTORY, REGISTRY)
        clear: True will remove non-Inherited ACEs from the ACL

    Returns (dict): A dictionary containing the results

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' win_dacl.enable_inheritance c:\temp directory
    """
    dc = daclConstants()
    objectType = dc.getObjectTypeBit(objectType)
    path = dc.processPath(path, objectType)

    return _set_dacl_inheritance(path, objectType, True, None, clear)


def disable_inheritance(path, objectType, copy=True):
    """
    Disable inheritance on an object

    Args:
        path: The path to the object
        objectType: The type of object (FILE, DIRECTORY, REGISTRY)
        copy: True will copy the Inherited ACEs to the DACL before disabling inheritance

    Returns (dict): A dictionary containing the results

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' win_dacl.disable_inheritance c:\temp directory
    """
    dc = daclConstants()
    objectType = dc.getObjectTypeBit(objectType)
    path = dc.processPath(path, objectType)

    return _set_dacl_inheritance(path, objectType, False, copy, None)


def check_inheritance(path, objectType, user=None):
    """
    Check a specified path to verify if inheritance is enabled

    Args:
        path: path of the registry key or file system object to check
        objectType: The type of object (FILE, DIRECTORY, REGISTRY)
        user: if provided, will consider only the ACEs for that user

    Returns (bool): 'Inheritance' of True/False

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' win_dacl.check_inheritance c:\temp directory <username>
    """

    ret = {"result": False, "Inheritance": False, "comment": ""}

    sidRet = _getUserSid(user)

    dc = daclConstants()
    objectType = dc.getObjectTypeBit(objectType)
    path = dc.processPath(path, objectType)

    try:
        sd = win32security.GetNamedSecurityInfo(
            path, objectType, win32security.DACL_SECURITY_INFORMATION
        )
        dacls = sd.GetSecurityDescriptorDacl()
    except Exception as e:  # pylint: disable=broad-except
        ret["result"] = False
        ret[
            "comment"
        ] = "Error obtaining the Security Descriptor or DACL of the path: {}.".format(e)
        return ret

    for counter in range(0, dacls.GetAceCount()):
        ace = dacls.GetAce(counter)
        if (ace[0][1] & win32security.INHERITED_ACE) == win32security.INHERITED_ACE:
            if not sidRet["sid"] or ace[2] == sidRet["sid"]:
                ret["Inheritance"] = True
                break

    ret["result"] = True
    return ret


def check_ace(
    path,
    objectType,
    user,
    permission=None,
    acetype=None,
    propagation=None,
    exactPermissionMatch=False,
):
    """
    Checks a path to verify the ACE (access control entry) specified exists

    Args:
        path:  path to the file/reg key
        objectType: The type of object (FILE, DIRECTORY, REGISTRY)
        user:  user that the ACL is for
        permission:  permission to test for (READ, FULLCONTROL, etc)
        acetype:  the type of ACE (ALLOW or DENY)
        propagation:  the propagation type of the ACE (FILES, FOLDERS, KEY, KEY&SUBKEYS, SUBKEYS, etc)
        exactPermissionMatch:  the ACL must match exactly, IE if READ is specified, the user must have READ exactly and not FULLCONTROL (which also has the READ permission obviously)

    Returns (dict): 'Exists' true if the ACE exists, false if it does not

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' win_dacl.check_ace c:\temp directory <username> fullcontrol
    """
    ret = {"result": False, "Exists": False, "comment": ""}

    dc = daclConstants()
    objectTypeBit = dc.getObjectTypeBit(objectType)
    path = dc.processPath(path, objectTypeBit)

    permission = permission.upper() if permission else None
    acetype = acetype.upper() if permission else None
    propagation = propagation.upper() if propagation else None

    permissionbit = (
        dc.getPermissionBit(objectTypeBit, permission) if permission else None
    )
    acetypebit = dc.getAceTypeBit(acetype) if acetype else None
    propagationbit = (
        dc.getPropagationBit(objectTypeBit, propagation) if propagation else None
    )

    sidRet = _getUserSid(user)
    if not sidRet["result"]:
        return sidRet

    dacls = _get_dacl(path, objectTypeBit)
    ret["result"] = True
    if dacls:
        for counter in range(0, dacls.GetAceCount()):
            ace = dacls.GetAce(counter)
            if ace[2] == sidRet["sid"]:
                if not acetypebit or ace[0][0] == acetypebit:
                    if (
                        not propagationbit
                        or (ace[0][1] & propagationbit) == propagationbit
                    ):
                        if not permissionbit:
                            ret["Exists"] = True
                            return ret
                        if exactPermissionMatch:
                            if ace[1] == permissionbit:
                                ret["Exists"] = True
                                return ret
                        else:
                            if (ace[1] & permissionbit) == permissionbit:
                                ret["Exists"] = True
                                return ret
    else:
        ret["comment"] = "No DACL found for object."
    return ret
